import logging
import os
import subprocess
import sys
import time
from collections.abc import Generator

from ruuvitag_sensor.adapters import BleCommunication
from ruuvitag_sensor.ruuvi_types import MacAndRawData, RawData

log = logging.getLogger(__name__)


class BleCommunicationNix(BleCommunication):
    """Bluetooth LE communication for Linux"""

    @staticmethod
    def start(bt_device=""):
        """
        Attributes:
           device (string): BLE device (default hci0)
        """
        # import ptyprocess here so as long as all implementations are in
        # the same file, all will work
        import ptyprocess  # noqa: PLC0415

        if not bt_device:
            bt_device = "hci0"

        is_root = os.getuid() == 0

        log.info("Start receiving broadcasts (device %s)", bt_device)
        DEVNULL = subprocess.DEVNULL

        def reset_ble_adapter():
            cmd = f"hciconfig {bt_device} reset"
            log.info("FYI: Calling a process%s: %s", "" if is_root else " with sudo", cmd)

            cmd = f"sudo {cmd}" if not is_root else cmd
            return subprocess.call(cmd, shell=True, stdout=DEVNULL)

        def start_with_retry(func, try_count, interval, msg):
            retcode = func()
            if retcode != 0 and try_count > 0:
                log.info(msg)
                time.sleep(interval)
                return start_with_retry(func, try_count - 1, interval + interval, msg)
            return retcode

        retcode = start_with_retry(reset_ble_adapter, 3, 1, "Problem with hciconfig reset. Retry reset.")

        if retcode != 0:
            log.info("Problem with hciconfig reset. Exit.")
            sys.exit(1)

        cmd = ["hcitool", "-i", bt_device, "lescan2", "--duplicates", "--passive"]
        log.info("FYI: Spawning process%s: %s", "" if is_root else " with sudo", " ".join(str(i) for i in cmd))

        if not is_root:
            cmd.insert(0, "sudo")
        hcitool = ptyprocess.PtyProcess.spawn(cmd)

        cmd = ["hcidump", "-i", bt_device, "--raw"]
        log.info("FYI: Spawning process%s: %s", "" if is_root else " with sudo", " ".join(str(i) for i in cmd))

        if not is_root:
            cmd.insert(0, "sudo")
        hcidump = ptyprocess.PtyProcess.spawn(cmd)

        return (hcitool, hcidump)

    @staticmethod
    def stop(hcitool, hcidump):
        log.info("Stop receiving broadcasts")
        hcitool.close()
        hcidump.close()

    @staticmethod
    def get_lines(hcidump):
        data = None
        try:
            while True:
                line = hcidump.readline().decode()
                if line == "":
                    # EOF reached
                    raise Exception("EOF received from hcidump")

                line = line.strip()
                log.debug("Read line from hcidump: %s", line)
                if line.startswith("> "):
                    log.debug("Yielding %s", data)
                    yield data
                    data = line[2:].replace(" ", "")
                elif line.startswith("< "):
                    data = None
                elif data:
                    data += line.replace(" ", "")
        except KeyboardInterrupt:
            return
        except Exception as ex:
            log.info(ex)
            return

    @staticmethod
    def get_data(blacklist: list[str] | None = None, bt_device: str = "") -> Generator[MacAndRawData, None, None]:
        procs = BleCommunicationNix.start(bt_device)
        data = None
        for line in BleCommunicationNix.get_lines(procs[1]):
            log.debug("Parsing line %s", line)
            try:
                # Make sure we're in upper case
                line = line.upper()  # noqa: PLW2901
                # We're interested in LE meta events, sent by Ruuvitags.
                # Those start with "043E", followed by a length byte.

                if not line.startswith("043E"):
                    log.debug("Not a LE meta packet")
                    continue

                # The third byte is the parameter length, this should cover
                # the length of the entire packet, minus the first three bytes.
                # Note that the data is in hex format, so use two chars per
                # byte
                plen = int(line[4:6], 16)
                if plen != (len(line) / 2) - 3:
                    log.debug("Invalid parameter length")
                    continue

                # The following two bytes should be "0201", indicating
                # 02  LE Advertising report
                # 01  1 report

                if line[6:10] != "0201":
                    log.debug("Not a Ruuvi advertisement")
                    continue

                # The next four bytes indicate whether the endpoint is
                # connectable or not, and whether the MAC address is random
                # or not. Different tags set different values here, so
                # ignore those.

                # The following 6 bytes are the MAC address of the sender,
                # in reverse order

                found_mac = line[14:26]
                reversed_mac = "".join(reversed([found_mac[i : i + 2] for i in range(0, len(found_mac), 2)]))
                mac = ":".join(a + b for a, b in zip(reversed_mac[::2], reversed_mac[1::2], strict=True))
                if blacklist and mac in blacklist:
                    log.debug("MAC blacklisted: %s", mac)
                    continue
                data = line[26:]
                log.debug("MAC: %s, data: %s", mac, data)
                yield (mac, data)
            except GeneratorExit:
                break
            except Exception:
                continue

        BleCommunicationNix.stop(procs[0], procs[1])

    @staticmethod
    def get_first_data(mac: str, bt_device: str = "") -> RawData:
        data = None
        data_iter = BleCommunicationNix.get_data([], bt_device)
        for d in data_iter:
            if mac == d[0]:
                log.info("Data found")
                data_iter.close()
                data = d[1]
                break

        return data or ""



import re

from ruuvitag_sensor.decoder import get_decoder
from ruuvitag_sensor.ruuvi import RuuviTagSensor
from ruuvitag_sensor.ruuvi_types import DataFormat, SensorData

mac_regex = "[0-9a-f]{2}([:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$"


class RuuviTagBase:
    """
    RuuviTag Sensors object
    """

    def __init__(self, mac: str, bt_device: str = ""):
        if not re.match(mac_regex, mac.lower()):
            raise ValueError(f"{mac} is not a valid MAC address")

        self._mac: str = mac
        self._state: dict | SensorData = {}
        self._data: str | None = None
        self._bt_device: str = bt_device

    @property
    def mac(self) -> str:
        return self._mac

    @property
    def state(self) -> dict | SensorData:
        return self._state

    def _handle_new_data_and_return_state(self, data_format: DataFormat, data: str | None) -> dict | SensorData:
        if data == self._data:
            return self._state

        self._data = data

        if self._data is None:
            self._state = {}
        elif data_format is not None:
            self._state = get_decoder(data_format).decode_data(self._data)  # type: ignore[assignment]

        return self._state


class RuuviTag(RuuviTagBase):
    def update(self) -> dict | SensorData:
        """
        Get latest data from the sensor and update own state.

        Returns:
            dict: Latest state
        """

        (data_format, raw_data) = RuuviTagSensor.get_first_raw_data(self._mac, self._bt_device)
        return self._handle_new_data_and_return_state(data_format, raw_data)


class RuuviTagAsync(RuuviTagBase):
    """
    NOTE: This class is not working on macOS
    """

    async def update(self) -> dict | SensorData:
        """
        Get latest data from the sensor and update own state.

        Returns:
            dict: Latest state
        """

        (data_format, raw_data) = await RuuviTagSensor.get_first_raw_data_async(self._mac, self._bt_device)
        return self._handle_new_data_and_return_state(data_format, raw_data)



import logging

from ruuvitag_sensor.ruuvi_types import DataFormatAndRawSensorData, RawSensorData

log = logging.getLogger(__name__)


class ShortDataError(Exception):
    pass


def _dechunk(raw: str) -> tuple[str, str]:
    """
    Given a BLE advertisement in hex format, interpret the first
    byte as a length byte, return the data indicated by the length
    byte, and the remainder of the data in a tuple.

    The length byte itself is not included in the length.

    If the length indicated is longer than the data, raise a ValueError
    """
    if len(raw) < 2:
        raise ShortDataError("Data too short")

    dlen = int(raw[:2], 16)
    if (dlen + 1) * 2 > len(raw):
        raise ShortDataError(f"Cannot read {dlen} bytes, data too short: {raw}")

    return (raw[2 : (dlen * 2) + 2], raw[(dlen * 2) + 2 :])


class DataFormats:
    """
    RuuviTag and Ruuvi Air broadcasted raw data handling for each data format
    """

    @staticmethod
    def convert_data(raw: str) -> DataFormatAndRawSensorData:  # noqa: PLR0911, PLR0912, C901
        """
        Validate that data is from RuuviTag or Ruuvi Air and get correct data part.

        There is a special case where this function will return
        None for the data format, and '' for the data. This indicates
        that we just heard an advertisement from a Ruuvi tag that
        doesn't contain any data, but was sent for discovery purposes
        (firmware 3.x does this).

        Returns:
            tuple (int, string): Data Format type and Sensor data
        """
        log.debug("Parsing advertisement data: %s", raw)

        try:
            # The data starts with a length byte, covering the data
            # length, minus the length byte itself. There might be additional
            # data at the end: the RSSI value
            data, rssi = _dechunk(raw)

            # The remaining data is a list of length:type:data chunks.
            # We look for a chunk with vendor specific data (type 0xff),
            # used by formats 3 and 5, or a chunk with service data (type 0x16)
            # used by formats 2 and 4.
            #
            # Firmware 3.x also sends advertisements that contain chunks
            # of type 0x09, followed by 'Ruuvi', in ASCII encoding.
            candidate = None
            while data != "":
                cdata, data = _dechunk(data)

                ctype = cdata[:2]
                log.debug("Found chunk of type %s: %s", ctype, cdata)

                # See if we found a potential candidate. Break the loop
                if ctype in ("FF", "16", "09"):
                    candidate = cdata
                    break
        except ShortDataError as ex:
            # Data might be from RuuviTag, but received data was invalid
            # e.g. it's possible that Bluetooth stack received only partial data
            # Set the format to None, and data to '', this allows the
            # caller to determine that we did indeed see a RuuviTag.
            log.debug("Error parsing advertisement data: %s", ex)
            return (None, "")
        except Exception:
            log.exception("Invalid advertisement data: %s", raw)
            return (None, None)

        if candidate is None:
            log.debug("No candidate found")
            return (None, None)

        log.debug("Found candidate %s", candidate)

        # Ruuvi advertisements start with FF9904 (for format 3, 5, 6 and E1),
        # or 16AAFE (for format 2 and 4).
        match candidate:
            # Data formats are ordered by priority, so the most common formats are at the top.
            case _ if candidate.startswith("FF990405"):
                return (5, (candidate[6:] + rssi) if rssi else candidate[6:])
            case _ if candidate.startswith("FF990406"):
                return (6, candidate[6:])
            case _ if candidate.lower().startswith("ff9904e1"):
                # Ruuvi Air sends E1 in lowercase, e.g. `FF9904e1...`
                return ("E1", candidate[6:].lower())
            case _ if candidate.startswith("FF990403"):
                return (3, candidate[6:])
            case _ if candidate.startswith("16AAFE"):
                # TODO: Check from raw data correct data format
                # Now this returns 2 also for Data Format 4
                url_data = DataFormats._get_data_format_2and4(DataFormats._parse_raw(raw))
                if url_data is not None:
                    return (2, url_data)
                return (None, None)
            case _ if candidate.startswith("095275757669"):
                # This is a RuuviTag, but this advertisement does not contain any data.
                # Set the format to None, and data to '', this allows the
                # caller to determine that we did indeed see a RuuviTag.
                return (None, "")
            case _:
                return (None, None)

    @staticmethod
    def _parse_raw(raw: str) -> str:
        return raw

    @staticmethod
    def _get_data_format_2and4(raw: str) -> RawSensorData:
        """
        Validate that data is from RuuviTag and is Data Format 2 or 4.
        Convert hexadecimal data to string.
        Encoded data part is after ruu.vi/#

        Returns:
            string: Encoded sensor data
        """
        try:
            base16_split = [raw[i : i + 2] for i in range(0, len(raw), 2)]
            selected_hexs = filter(lambda x: int(x, 16) < 128, base16_split)
            characters = [chr(int(c, 16)) for c in selected_hexs]
            data = "".join(characters)

            # take only part after ruu.vi/#
            index = data.find("ruu.vi/#")
            if index > -1:
                return data[(index + 8) :]

            return None
        except Exception:
            return None

    @staticmethod
    def _get_data_format_3(raw: str) -> RawSensorData:
        """
        Validate that data is from RuuviTag and is Data Format 3

        Returns:
            string: Sensor data
        """
        # Search of FF990403 (Manufacturer Specific Data (FF) /
        # Ruuvi Innovations ltd (9904) / Format 3 (03))
        try:
            if "FF990403" not in raw:
                return None

            payload_start = raw.index("FF990403") + 6
            return raw[payload_start:]
        except Exception:
            return None

    @staticmethod
    def _get_data_format_5(raw: str) -> RawSensorData:
        """
        Validate that data is from RuuviTag and is Data Format 5

        Returns:
            string: Sensor data
        """
        # Search of FF990405 (Manufacturer Specific Data (FF) /
        # Ruuvi Innovations ltd (9904) / Format 5 (05))
        try:
            if "FF990405" not in raw:
                return None

            payload_start = raw.index("FF990405") + 6
            return raw[payload_start:]
        except Exception:
            return None

    @staticmethod
    def _get_data_format_e1(raw: str) -> RawSensorData:
        """
        Validate that data is from Ruuvi Air and is Data Format E1

        Returns:
            string: Sensor data
        """
        # Search of ff9904e1 (Manufacturer Specific Data (FF) /
        # Ruuvi Innovations ltd / Data format E1 (Extended v1)
        try:
            raw = raw.lower()
            if "ff9904e1" not in raw:
                return None

            payload_start = raw.index("ff9904e1") + 6
            return raw[payload_start:]
        except Exception:
            return None



import asyncio
import logging
import time
from collections.abc import AsyncGenerator, Callable, Generator
from datetime import datetime
from multiprocessing import Manager
from multiprocessing.managers import ListProxy
from warnings import warn

from ruuvitag_sensor.adapters import get_ble_adapter, throw_if_not_async_adapter, throw_if_not_sync_adapter
from ruuvitag_sensor.data_formats import DataFormats
from ruuvitag_sensor.decoder import AirHistoryDecoder, HistoryDecoder, get_decoder, parse_mac
from ruuvitag_sensor.ruuvi_types import (
    DataFormatAndRawSensorData,
    DeviceType,
    Mac,
    MacAndRawData,
    MacAndSensorData,
    SensorAirHistoryData,
    SensorData,
    SensorHistoryData,
)

log = logging.getLogger(__name__)
ble = get_ble_adapter()


class RunFlag:
    """
    Wrapper for boolean run flag

    Attributes:
        running (bool): Defines if the function should continue execution
    """

    running = True


class RuuviTagSensor:
    """
    RuuviTag communication functionality
    """

    @staticmethod
    def get_first_raw_data(mac: str, bt_device: str = "") -> DataFormatAndRawSensorData:
        """
        Get raw data for selected RuuviTag. This method is intended to be used only by
        RuuviTag-class.

        Args:
            mac (string): MAC address
            bt_device (string): Bluetooth device id
        Returns:
            tuple (int, string): Data Format type and raw Sensor data
        """
        throw_if_not_sync_adapter(ble)

        raw = ble.get_first_data(mac, bt_device)
        return DataFormats.convert_data(raw)

    @staticmethod
    async def get_first_raw_data_async(mac: str, bt_device: str = "") -> DataFormatAndRawSensorData:
        """
        Get raw data for selected RuuviTag. This method is intended to be used only by
        RuuviTagAsync-class.

        NOTE: This method is not working on macOS

        Args:
            mac (string): MAC address
            bt_device (string): Bluetooth device id
        Returns:
            tuple (int, string): Data Format type and raw Sensor data
        """
        throw_if_not_async_adapter(ble)

        raw = await ble.get_first_data(mac, bt_device)
        return DataFormats.convert_data(raw)

    @staticmethod
    def find_ruuvitags(bt_device: str = "") -> dict[Mac, SensorData]:
        """
        CLI helper function.

        Find all RuuviTags. Function will print the MAC and the state of the sensors when found.
        Function will execute as long as it is stopped. Stop execution with Ctrl+C.

        Returns:
            dict: MAC and state of found sensors
        """
        throw_if_not_sync_adapter(ble)

        log.info("Finding RuuviTags. Stop with Ctrl+C.")

        data: dict[str, SensorData] = {}
        for new_data in RuuviTagSensor._get_ruuvitag_data(bt_device=bt_device):
            mac, sensor_data = new_data
            if not mac or mac in data:
                continue
            data[mac] = sensor_data
            log.info(mac)
            log.info(sensor_data)

        return data

    @staticmethod
    async def find_ruuvitags_async(bt_device: str = "") -> dict[Mac, MacAndSensorData]:
        """
        CLI helper function.

        Find all RuuviTags. Function will print the MAC and the state of the sensors when found.
        Function will execute as long as it is stopped. Stop execution with Ctrl+C.

        Returns:
            dict: MAC and state of found sensors
        """
        throw_if_not_async_adapter(ble)

        log.info("Finding RuuviTags. Stop with Ctrl+C.")

        data: dict[Mac, MacAndSensorData] = {}
        mac_blacklist = Manager().list()
        data_iter = ble.get_data(mac_blacklist, bt_device)

        try:
            async for new_data in data_iter:
                if new_data[0] in data:
                    continue

                parsed_data = RuuviTagSensor._parse_data(new_data, mac_blacklist)
                if parsed_data:
                    data[new_data[0]] = parsed_data
                    log.info(new_data[0])
                    log.info(parsed_data)
        finally:
            await data_iter.aclose()

        return data

    @staticmethod
    def get_data_for_sensors(
        macs: list[str] | None = None, search_duration_sec: int = 5, bt_device: str = ""
    ) -> dict[Mac, SensorData]:
        """
        Get latest data for RuuviTag and Ruuvi Air sensors in the MAC address list.

        Args:
            macs (array): MAC addresses
            search_duration_sec (int): Search duration in seconds. Default 5
            bt_device (string): Bluetooth device id
        Returns:
            dict: MAC and state of found sensors
        """
        if macs is None:
            macs = []

        throw_if_not_sync_adapter(ble)

        log.info("Get latest data for sensors. Stop with Ctrl+C.")
        log.info("Stops automatically in %ss", search_duration_sec)
        log.info("MACs: %s", macs)

        data: dict[Mac, SensorData] = {}

        for new_data in RuuviTagSensor._get_ruuvitag_data(macs, search_duration_sec, bt_device=bt_device):
            mac, sensor_data = new_data
            data[mac] = sensor_data

        return data

    @staticmethod
    async def get_data_for_sensors_async(
        macs: list[str] | None = None, search_duration_sec: int = 5, bt_device: str = ""
    ) -> dict[Mac, SensorData]:
        """
        Get latest data for RuuviTag and Ruuvi Air sensors in the MAC address list.

        Args:
            macs (array): MAC addresses
            search_duration_sec (int): Search duration in seconds. Default 5
            bt_device (string): Bluetooth device id
        Returns:
            dict: MAC and state of found sensors
        """
        if macs is None:
            macs = []

        throw_if_not_async_adapter(ble)

        log.info("Get latest data for sensors. Stop with Ctrl+C.")
        log.info("Stops automatically in %ss", search_duration_sec)
        log.info("MACs: %s", macs)

        data: dict[Mac, SensorData] = {}
        start_time = time.time()

        data_iter = RuuviTagSensor.get_data_async(macs, bt_device)

        try:
            async for new_data in data_iter:
                mac, sensor_data = new_data
                data[mac] = sensor_data
                if search_duration_sec and time.time() - start_time > search_duration_sec:
                    break
        finally:
            await data_iter.aclose()

        return data

    @staticmethod
    async def get_data_async(
        macs: list[str] | None = None, bt_device: str = ""
    ) -> AsyncGenerator[MacAndSensorData, None]:
        """
        Get data for all RuuviTag and Ruuvi Air sensors or sensors in the MAC's list.

        Args:
            macs (list): MAC addresses
            bt_device (string): Bluetooth device id
        Returns:
            AsyncGenerator: MAC and State of sensor data (tuple)
        """
        if macs is None:
            macs = []

        throw_if_not_async_adapter(ble)

        mac_blacklist = Manager().list()
        data_iter = ble.get_data(mac_blacklist, bt_device)

        try:
            async for ble_data in data_iter:
                data = RuuviTagSensor._parse_data(ble_data, mac_blacklist, macs)

                # Check MAC whitelist if advertised MAC available
                if ble_data[0] and macs and ble_data[0] not in macs:
                    log.debug("MAC not whitelisted: %s", ble_data[0])
                    continue

                if data:
                    yield data
        finally:
            await data_iter.aclose()

    @staticmethod
    def get_data(
        callback: Callable[[MacAndSensorData], None],
        macs: list[str] | None = None,
        run_flag: RunFlag | None = None,
        bt_device: str = "",
    ) -> None:
        """
        Get data for all RuuviTag and Ruuvi Air sensors or sensors in the MAC's list.

        Args:
            callback (func): callback function to be called when new data is received
            macs (list): MAC addresses
            run_flag (object): RunFlag object. Function executes while run_flag.running
            bt_device (string): Bluetooth device id
        """
        if macs is None:
            macs = []
        if run_flag is None:
            run_flag = RunFlag()

        throw_if_not_sync_adapter(ble)

        log.info("Get latest data for sensors. Stop with Ctrl+C.")
        log.info("MACs: %s", macs)

        for new_data in RuuviTagSensor._get_ruuvitag_data(macs, None, run_flag, bt_device):
            callback(new_data)

    @staticmethod
    def get_datas(
        callback: Callable[[MacAndSensorData], None],
        macs: list[str] | None = None,
        run_flag: RunFlag | None = None,
        bt_device: str = "",
    ) -> None:
        """
        DEPRECATED
        This method will be removed in a future version.
        Use get_data-method instead.
        """
        if macs is None:
            macs = []
        if run_flag is None:
            run_flag = RunFlag()

        warn("This method will be removed in a future version, use get_data() instead", FutureWarning, stacklevel=2)
        throw_if_not_sync_adapter(ble)
        return RuuviTagSensor.get_data(callback, macs, run_flag, bt_device)

    @staticmethod
    def _get_ruuvitag_data(
        macs: list[str] | None = None,
        search_duration_sec: int | None = None,
        run_flag: RunFlag | None = None,
        bt_device: str = "",
    ) -> Generator[MacAndSensorData, None, None]:
        """
        Get data from BluetoothCommunication and handle data encoding.

        Args:
            macs (list): MAC addresses. Default empty list
            search_duration_sec (int): Search duration in seconds. Default None
            run_flag (object): RunFlag object. Function executes while run_flag.running.
                               Default new RunFlag
            bt_device (string): Bluetooth device id
        Yields:
            tuple: MAC and State of sensor data
        """
        if macs is None:
            macs = []
        if run_flag is None:
            run_flag = RunFlag()

        mac_blacklist = Manager().list()
        start_time = time.time()
        data_iter = ble.get_data(mac_blacklist, bt_device)

        for ble_data in data_iter:
            if search_duration_sec and time.time() - start_time > search_duration_sec:
                data_iter.close()
                break
            if not run_flag.running:
                data_iter.close()
                break
            if ble_data[0] and macs and ble_data[0] not in macs:
                log.debug("MAC not whitelisted: %s", ble_data[0])
                continue

            data = RuuviTagSensor._parse_data(ble_data, mac_blacklist, macs)
            if data:
                yield data

    @staticmethod
    def _parse_data(
        ble_data: MacAndRawData, mac_blacklist: ListProxy, allowed_macs: list[str] | None = None
    ) -> MacAndSensorData | None:
        if allowed_macs is None:
            allowed_macs = []
        (mac, payload) = ble_data
        (data_format, data) = DataFormats.convert_data(payload)

        # Check that encoded data is valid RuuviTag data and it is sensor data
        # If data is not valid RuuviTag data add MAC to blacklist if MAC is available
        if data is None:
            if mac:
                log.debug("Blacklisting MAC %s", mac)
                mac_blacklist.append(mac)
            return None

        if data_format is None:
            # Whatever we heard was from a Ruuvitag, but did not contain
            # any measurements. Ignore this.
            return None

        decoded = get_decoder(data_format).decode_data(data)
        if decoded is None:
            log.error("Decoded data is null. MAC: %s - Raw: %s", mac, payload)
            return None

        # If advertised MAC is missing, try to parse it from the payload
        mac_to_send = (
            mac
            if mac
            else parse_mac(data_format, decoded["mac"])  # type: ignore[typeddict-item]
            if "mac" in decoded and decoded["mac"] is not None  # type: ignore[typeddict-item]
            else ""
        )

        # Check whitelist using MAC from decoded data if advertised MAC is not available
        if allowed_macs and mac_to_send not in allowed_macs:
            log.debug("MAC not whitelisted: %s", mac_to_send)
            return None

        return (mac_to_send, decoded)

    @staticmethod
    async def get_history_async(
        mac: str,
        start_time: datetime | None = None,
        max_items: int | None = None,
        device_type: DeviceType = "ruuvitag",
    ) -> AsyncGenerator[SensorHistoryData | SensorAirHistoryData, None]:
        """
        Get history data from a RuuviTag or Ruuvi Air as an async stream.

        For RuuviTag: Each history entry contains one measurement type (temperature, humidity, or pressure)
        with Unix timestamp (integer). Requires firmware version 3.30.0 or newer.

        For Ruuvi Air: Each history entry contains all sensor measurements (temperature, humidity, pressure,
        PM values, CO₂, VOC, NOx) with Unix timestamp (integer).

        Args:
            mac (str): MAC address or UUID of the device. On macOS use UUID instead.
            start_time (Optional[datetime]): If provided, only get data from this time onwards. Time should be in UTC.
            max_items (Optional[int]): Maximum number of history entries to fetch. If None, gets all available data
            device_type (DeviceType): Device type - "ruuvi_air" for Ruuvi Air,
                "ruuvitag" for RuuviTag (default: "ruuvitag")

        Yields:
            SensorHistoryData or SensorAirHistoryData: Individual history measurements.
                Note: Ruuvi Air may transmit multiple records per BLE packet, but this method yields one decoded
                record at a time.

        Raises:
            RuntimeError: If connection fails or device doesn't support history
        """
        throw_if_not_async_adapter(ble)

        if device_type not in {"ruuvitag", "ruuvi_air"}:
            raise ValueError(f"Invalid device_type: {device_type}. Must be 'ruuvitag' or 'ruuvi_air'")

        decoder = HistoryDecoder() if device_type == "ruuvitag" else AirHistoryDecoder()

        try:
            data_iter = ble.get_history_data(mac, start_time, max_items, device_type=device_type)
            async for data in data_iter:
                if decoded := decoder.decode_data(data):
                    if isinstance(decoded, list):
                        for record in decoded:
                            yield record
                    else:
                        yield decoded
        finally:
            await data_iter.aclose()

    @staticmethod
    async def download_history(
        mac: str,
        start_time: datetime | None = None,
        timeout: int = 300,
        max_items: int | None = None,
        device_type: DeviceType = "ruuvitag",
    ) -> list[SensorHistoryData | SensorAirHistoryData]:
        """
        Download complete history data from a RuuviTag or Ruuvi Air.
        This method collects all history entries and returns them as a list.

        For RuuviTag: Each history entry contains one measurement type (temperature, humidity, or pressure)
        with Unix timestamp (integer). Requires firmware version 3.30.0 or newer.
        Note: The RuuviTag sends each measurement type as separate entries.
        If you need to combine measurements by timestamp, you'll need to post-process the data.

        For Ruuvi Air: Each history entry contains all sensor measurements (temperature, humidity, pressure,
        PM values, CO₂, VOC, NOx) with Unix timestamp (integer).

        Args:
            mac (str): MAC address or UUID of the device. On macOS use UUID instead.
            start_time (Optional[datetime]): If provided, only get data from this time onwards. Time should be in UTC.
            timeout (int): Maximum time in seconds to wait for history download (default: 300)
            max_items (Optional[int]): Maximum number of history entries to fetch. If None, gets all available data
            device_type (DeviceType): Device type - "ruuvi_air" for Ruuvi Air,
                "ruuvitag" for RuuviTag (default: "ruuvitag")

        Returns:
            List[SensorHistoryData] or List[SensorAirHistoryData]: List of historical measurements

        Raises:
            RuntimeError: If connection fails or device doesn't support history
            TimeoutError: If download takes longer than timeout
        """
        throw_if_not_async_adapter(ble)

        if device_type not in {"ruuvitag", "ruuvi_air"}:
            raise ValueError(f"Invalid device_type: {device_type}. Must be 'ruuvitag' or 'ruuvi_air'")

        decoder = HistoryDecoder() if device_type == "ruuvitag" else AirHistoryDecoder()

        try:
            history_data: list[SensorHistoryData | SensorAirHistoryData] = []

            async def collect_history():
                async for data in ble.get_history_data(mac, start_time, max_items, device_type=device_type):
                    if decoded := decoder.decode_data(data):
                        # HistoryDecoder returns single record or None, AirHistoryDecoder returns list
                        history_data.extend(decoded if isinstance(decoded, list) else [decoded])

            await asyncio.wait_for(collect_history(), timeout=timeout)
            return history_data

        except asyncio.TimeoutError as err:
            raise TimeoutError(f"History download timed out after {timeout} seconds") from err
        except Exception as e:
            raise RuntimeError(f"Failed to download history: {e!s}") from e



from __future__ import annotations

import logging
import math
import struct

from ruuvitag_sensor.ruuvi_types import ByteData, SensorData3

log = logging.getLogger(__name__)


class Df3Decoder:
    """
    Decodes data from RuuviTag with Data Format 3
    Protocol specification:
    https://github.com/ruuvi/ruuvi-sensor-protocols
    """

    def _get_temperature(self, data: ByteData) -> float:
        """Return temperature in celsius"""

        # The temperature is in two fields, one for the integer part,
        # one for the fraction
        #
        # The integer part was decoded as a signed two's complement number,
        # but this isn't how it's really stored. The MSB is a sign, the lower
        # 7 bits are the unsigned temperature value.
        #
        # To convert from the decoded value we have to add 128 and then negate,
        # if the decoded value was negative
        frac = data[3] / 100
        if data[2] < 0:
            return -(data[2] + 128 + frac)

        return data[2] + frac

    def _get_humidity(self, data: ByteData) -> float:
        """Return humidity %"""
        return data[1] * 0.5

    def _get_pressure(self, data: ByteData) -> float:
        """Return air pressure hPa"""
        return (data[4] + 50000) / 100

    def _get_acceleration(self, data: ByteData) -> tuple[int, int, int]:
        """Return acceleration mG"""
        return data[5:8]  # type: ignore

    def _get_battery(self, data: ByteData) -> int:
        """Return battery mV"""
        return data[8]

    def decode_data(self, data: str) -> SensorData3 | None:
        """
        Decode sensor data.

        Returns:
            dict: Sensor values
        """
        try:
            byte_data: ByteData = struct.unpack(">BBbBHhhhH", bytearray.fromhex(data[:28]))
            acc_x, acc_y, acc_z = self._get_acceleration(byte_data)
            return {
                "data_format": 3,
                "humidity": self._get_humidity(byte_data),
                "temperature": self._get_temperature(byte_data),
                "pressure": self._get_pressure(byte_data),
                "acceleration": math.sqrt(acc_x * acc_x + acc_y * acc_y + acc_z * acc_z),
                "acceleration_x": acc_x,
                "acceleration_y": acc_y,
                "acceleration_z": acc_z,
                "battery": self._get_battery(byte_data),
            }
        except Exception:
            log.exception("Value: %s not valid", data)
            return None



from ruuvitag_sensor.adapters.bleak_ble import BleCommunicationBleak, HistoryNotificationAction


class TestHistoryNotificationProcessing:
    def test_ignore_heartbeat_data(self):
        """Test that heartbeat data (starting with 0x05) is ignored."""
        adapter = BleCommunicationBleak()
        # Heartbeat data starts with 0x05
        heartbeat_data = bytearray([0x05, 0x01, 0x02, 0x03, 0x04])

        action, data = adapter._process_history_notification(heartbeat_data)

        assert action == HistoryNotificationAction.IGNORE
        assert data is None

    def test_ignore_empty_heartbeat(self):
        """Test that empty heartbeat data is handled."""
        adapter = BleCommunicationBleak()
        # Empty data with first byte 0x05
        heartbeat_data = bytearray([0x05])

        action, data = adapter._process_history_notification(heartbeat_data)

        assert action == HistoryNotificationAction.IGNORE
        assert data is None

    def test_detect_end_marker(self):
        """Test detection of end-of-logs marker.

        Reference: https://docs.ruuvi.com/communication/bluetooth-connection/nordic-uart-service-nus/log-read
        End marker: 0x3A 3A 10 0xFFFFFFFF FFFFFFFF (all 0xFF)
        """
        adapter = BleCommunicationBleak()
        # End marker from official docs: 0x3A 3A 10 with payload all 0xFF
        end_marker = bytearray.fromhex("3a3a10ffffffffffffffff")

        action, data = adapter._process_history_notification(end_marker)

        assert action == HistoryNotificationAction.END
        assert data == end_marker

    def test_detect_end_marker_variations(self):
        """Test end marker detection with different lengths."""
        adapter = BleCommunicationBleak()
        # End marker with minimum length (3 bytes + all 0xFF)
        end_marker_short = bytearray([0x3A, 0x3A, 0x10, 0xFF, 0xFF, 0xFF])

        action, data = adapter._process_history_notification(end_marker_short)

        assert action == HistoryNotificationAction.END
        assert data == end_marker_short

    def test_detect_error_packet(self):
        """Test detection of error packet.

        Reference: https://docs.ruuvi.com/communication/bluetooth-connection/nordic-uart-service-nus/log-read
        Error message: Header type 0xF0 with payload 0xFF, e.g. 0x30 30 F0 FF FF FF FF FF FF FF FF
        """
        adapter = BleCommunicationBleak()
        # Error packet from official docs: 0x30 30 F0 FF FF FF FF FF FF FF FF
        error_data = bytearray.fromhex("3030f0ffffffffffffffff")

        action, data = adapter._process_history_notification(error_data)

        assert action == HistoryNotificationAction.ERROR
        assert data == error_data

    def test_error_packet_different_first_bytes(self):
        """Test error packet detection with different first bytes."""
        adapter = BleCommunicationBleak()
        # Error packet can have different first bytes, but data[2] == 0xF0
        error_data = bytearray([0x3A, 0x30, 0xF0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF])

        action, data = adapter._process_history_notification(error_data)

        assert action == HistoryNotificationAction.ERROR
        assert data == error_data

    def test_error_packet_too_short(self):
        """Test that error packet detection requires at least 11 bytes."""
        adapter = BleCommunicationBleak()
        # Too short to be an error packet (needs at least 11 bytes)
        short_data = bytearray([0x30, 0x30, 0xF0, 0xFF, 0xFF])

        action, data = adapter._process_history_notification(short_data)

        # Should be treated as normal data, not error
        assert action == HistoryNotificationAction.DATA
        assert data == short_data

    def test_normal_temperature_data(self):
        """Test processing of normal temperature history data."""
        adapter = BleCommunicationBleak()
        # Normal temperature entry from official docs: 0x3A 30 10 5D57FEAD 000000098D
        temp_data = bytearray.fromhex("3a30105d57fead000000098d")

        action, data = adapter._process_history_notification(temp_data)

        assert action == HistoryNotificationAction.DATA
        assert data == temp_data

    def test_normal_humidity_data(self):
        """Test processing of normal humidity history data."""
        adapter = BleCommunicationBleak()
        # Normal humidity entry from official docs: 0x3A 31 10 5D57FEAD 000000098D
        humidity_data = bytearray.fromhex("3a31105d57fead000000098d")

        action, data = adapter._process_history_notification(humidity_data)

        assert action == HistoryNotificationAction.DATA
        assert data == humidity_data

    def test_normal_pressure_data(self):
        """Test processing of normal pressure history data."""
        adapter = BleCommunicationBleak()
        # Normal pressure entry from official docs: 0x3A 32 10 5D57FEAD 000000098D
        pressure_data = bytearray.fromhex("3a32105d57fead000000098d")

        action, data = adapter._process_history_notification(pressure_data)

        assert action == HistoryNotificationAction.DATA
        assert data == pressure_data

    def test_empty_data(self):
        """Test handling of empty data."""
        adapter = BleCommunicationBleak()
        empty_data = bytearray()

        action, data = adapter._process_history_notification(empty_data)

        # Empty data should be treated as normal data (will be filtered by decoder)
        assert action == HistoryNotificationAction.DATA
        assert data == empty_data

    def test_short_data(self):
        """Test handling of short data (less than 3 bytes)."""
        adapter = BleCommunicationBleak()
        short_data = bytearray([0x3A, 0x30])

        action, data = adapter._process_history_notification(short_data)

        # Short data should be treated as normal data
        assert action == HistoryNotificationAction.DATA
        assert data == short_data

    def test_data_not_end_marker_if_not_all_ff(self):
        """Test that data is not treated as end marker if bytes after index 3 are not all 0xFF."""
        adapter = BleCommunicationBleak()
        # Has 0x3A 0x3A 0x10 but not all 0xFF after
        data = bytearray([0x3A, 0x3A, 0x10, 0xFF, 0xFF, 0x00, 0xFF])

        action, data_result = adapter._process_history_notification(data)

        assert action == HistoryNotificationAction.DATA
        assert data_result == data

    def test_data_not_error_if_f0_not_at_index_2(self):
        """Test that data is not treated as error if 0xF0 is not at index 2."""
        adapter = BleCommunicationBleak()
        # Has 0xF0 but not at index 2
        data = bytearray([0x30, 0xF0, 0x30, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF])

        action, data_result = adapter._process_history_notification(data)

        assert action == HistoryNotificationAction.DATA
        assert data_result == data

    def test_multiple_notifications_sequence(self):
        """Test processing a sequence of different notification types."""
        adapter = BleCommunicationBleak()

        # Heartbeat (should be ignored)
        action, data = adapter._process_history_notification(bytearray([0x05, 0x01]))
        assert action == HistoryNotificationAction.IGNORE
        assert data is None

        # Normal temperature data
        temp_data = bytearray.fromhex("3a30105d57fead000000098d")
        action, data = adapter._process_history_notification(temp_data)
        assert action == HistoryNotificationAction.DATA
        assert data == temp_data

        # Normal humidity data
        humidity_data = bytearray.fromhex("3a31105d57fead000000098d")
        action, data = adapter._process_history_notification(humidity_data)
        assert action == HistoryNotificationAction.DATA
        assert data == humidity_data

        # End marker
        end_marker = bytearray.fromhex("3a3a10ffffffffffffffff")
        action, data = adapter._process_history_notification(end_marker)
        assert action == HistoryNotificationAction.END
        assert data == end_marker



"""
Module level logging configuration for ruuvitag_sensor package.

This module provides:
1. A root logger for the package with default ERROR level file logging
2. A function to enable console output, primarily for CLI usage

Note: Applications using this package as a library should configure their own logging
rather than relying on this module's configuration.
"""

import logging

# Create the package's root logger
log = logging.getLogger("ruuvitag_sensor")
log.setLevel(logging.INFO)

# Configure file logging for errors
file_handler = logging.FileHandler("ruuvitag_sensor.log")
file_handler.setLevel(logging.ERROR)

# Set up a standard logging format with timestamp, logger name, level and message
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
file_handler.setFormatter(formatter)
log.addHandler(file_handler)


def enable_console(level: int = logging.INFO) -> None:
    """Enable console logging for the package.

    This function is primarily intended for command-line usage of the package.
    If the requested level is DEBUG, it will also set the root logger's level to DEBUG.
    The function ensures only one console handler is added.

    Args:
        level: The logging level for console output. Defaults to INFO.
    """
    if level < logging.INFO:
        log.setLevel(level)

    if len(log.handlers) != 2:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(level)
        log.addHandler(console_handler)



import logging

import Path

from ruuvitag_sensor.adapters.nix_hci import BleCommunicationNix

log = logging.getLogger(__name__)


class BleCommunicationNixFile(BleCommunicationNix):
    """
    Emulate HCI communication through hcidump by reading hcidump
    output from a file
    """

    @staticmethod
    def start(bt_device=""):
        """
        Attributes:
           device (string): BLE device (default hci0).
           This is interpreted as a file to open
        """
        log.info("Start reading from file %s", bt_device)
        handle = Path.open(bt_device, "rb")

        return (None, handle)

    @staticmethod
    def stop(_hcitool, hcidump):
        log.info("Close file")
        hcidump.close()



from __future__ import annotations

import logging
import math
import struct

from ruuvitag_sensor.ruuvi_types import ByteData, SensorData5

log = logging.getLogger(__name__)


class Df5Decoder:
    """
    Decodes data from RuuviTag with Data Format 5
    Protocol specification:
    https://github.com/ruuvi/ruuvi-sensor-protocols
    """

    def _get_temperature(self, data: ByteData) -> float | None:
        """Return temperature in celsius"""
        if data[1] == -32768:
            return None

        return round(data[1] / 200, 2)

    def _get_humidity(self, data: ByteData) -> float | None:
        """Return humidity %"""
        if data[2] == 65535:
            return None

        return round(data[2] / 400, 2)

    def _get_pressure(self, data: ByteData) -> float | None:
        """Return air pressure hPa"""
        if data[3] == 0xFFFF:
            return None

        return round((data[3] + 50000) / 100, 2)

    def _get_acceleration(self, data: ByteData) -> tuple[None, None, None] | tuple[int, int, int]:
        """Return acceleration mG"""
        if data[4] == -32768 or data[5] == -32768 or data[6] == -32768:
            return (None, None, None)

        return data[4:7]  # type: ignore

    def _get_powerinfo(self, data: ByteData) -> tuple[int, int]:
        """Return battery voltage and tx power"""
        battery_voltage = data[7] >> 5
        tx_power = data[7] & 0x001F

        return (battery_voltage, tx_power)

    def _get_battery(self, data: ByteData) -> int | None:
        """Return battery mV"""
        battery_voltage = self._get_powerinfo(data)[0]
        if battery_voltage == 0b11111111111:
            return None

        return battery_voltage + 1600

    def _get_txpower(self, data: ByteData) -> int | None:
        """Return transmit power"""
        tx_power = self._get_powerinfo(data)[1]
        if tx_power == 0b11111:
            return None

        return -40 + (tx_power * 2)

    def _get_movementcounter(self, data: ByteData) -> int:
        return data[8]

    def _get_measurementsequencenumber(self, data: ByteData) -> int:
        return data[9]

    def _get_mac(self, data: ByteData):
        return "".join(f"{x:02x}" for x in data[10:])

    def _get_rssi(self, rssi_byte: str) -> int:
        """Return RSSI value in dBm."""
        rssi = int(rssi_byte, 16)
        if rssi > 127:
            rssi = (256 - rssi) * -1
        return rssi

    def decode_data(self, data: str) -> SensorData5 | None:
        """
        Decode sensor data.

        Returns:
            dict: Sensor values
        """
        try:
            byte_data: ByteData = struct.unpack(">BhHHhhhHBH6B", bytearray.fromhex(data[:48]))
            rssi = data[48:]

            acc_x, acc_y, acc_z = self._get_acceleration(byte_data)
            acc = (
                math.sqrt(acc_x * acc_x + acc_y * acc_y + acc_z * acc_z)
                if acc_x is not None and acc_y is not None and acc_z is not None
                else None
            )

            # NOTE: Value parsing methods can return None, but it shouldn't happen with the
            # production firmware. Therefore properties are not optional on SensorData-type.

            return {
                "data_format": 5,
                "humidity": self._get_humidity(byte_data),  # type: ignore
                "temperature": self._get_temperature(byte_data),  # type: ignore
                "pressure": self._get_pressure(byte_data),  # type: ignore
                "acceleration": acc,  # type: ignore
                "acceleration_x": acc_x,  # type: ignore
                "acceleration_y": acc_y,  # type: ignore
                "acceleration_z": acc_z,  # type: ignore
                "tx_power": self._get_txpower(byte_data),  # type: ignore
                "battery": self._get_battery(byte_data),  # type: ignore
                "movement_counter": self._get_movementcounter(byte_data),
                "measurement_sequence_number": self._get_measurementsequencenumber(byte_data),
                "mac": self._get_mac(byte_data),
                "rssi": self._get_rssi(rssi) if rssi else None,
            }
        except Exception:
            log.exception("Value: %s not valid", data)
            return None



import abc
import os
from collections.abc import AsyncGenerator, Generator

from ruuvitag_sensor.ruuvi_types import MacAndRawData, RawData

# Disable imports at the top-level of a file
# ruff: noqa: PLC0415


def get_ble_adapter():
    forced_ble_adapter = os.environ.get("RUUVI_BLE_ADAPTER", "").lower()
    use_ruuvi_nix_from_file = "RUUVI_NIX_FROMFILE" in os.environ
    is_ci_env = "CI" in os.environ

    if forced_ble_adapter:
        if "bleak" in forced_ble_adapter:
            from ruuvitag_sensor.adapters.bleak_ble import BleCommunicationBleak

            return BleCommunicationBleak()
        if "bleson" in forced_ble_adapter:
            from ruuvitag_sensor.adapters.bleson import BleCommunicationBleson

            return BleCommunicationBleson()
        if "bluez" in forced_ble_adapter:
            from ruuvitag_sensor.adapters.nix_hci import BleCommunicationNix

            return BleCommunicationNix()

        raise RuntimeError(f"Unknown BLE adapter: {forced_ble_adapter}")

    if use_ruuvi_nix_from_file:
        # Emulate BleCommunicationNix by reading hcidump data from a file
        from ruuvitag_sensor.adapters.nix_hci_file import BleCommunicationNixFile

        return BleCommunicationNixFile()

    if is_ci_env:
        # Use BleCommunicationDummy for CI as it can't use Bleak/BlueZ
        from ruuvitag_sensor.adapters.dummy import BleCommunicationDummy

        return BleCommunicationDummy()

    # Bleak is default adapter for all platforms
    from ruuvitag_sensor.adapters.bleak_ble import BleCommunicationBleak

    return BleCommunicationBleak()


def is_async_adapter(ble: object):
    return issubclass(type(ble), BleCommunicationAsync)


def is_async_from_env():
    return "bleak" in os.environ.get("RUUVI_BLE_ADAPTER", "").lower()


def throw_if_not_sync_adapter(ble: object):
    if is_async_adapter(ble):
        raise RuntimeError("Sync BLE adapter required")


def throw_if_not_async_adapter(ble: object):
    if not is_async_adapter(ble):
        raise RuntimeError("Async BLE adapter required")


class BleCommunication:
    """Bluetooth LE communication"""

    __metaclass__ = abc.ABCMeta

    @staticmethod
    @abc.abstractmethod
    def get_first_data(mac: str, bt_device: str = "") -> RawData:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_data(blacklist: list[str] | None = None, bt_device: str = "") -> Generator[MacAndRawData, None, None]:
        pass


class BleCommunicationAsync:
    """Asynchronous Bluetooth LE communication"""

    __metaclass__ = abc.ABCMeta

    @staticmethod
    @abc.abstractmethod
    async def get_first_data(mac: str, bt_device: str = "") -> RawData:
        pass

    @staticmethod
    @abc.abstractmethod
    async def get_data(blacklist: list[str] | None = None, bt_device: str = "") -> AsyncGenerator[MacAndRawData, None]:
        raise NotImplementedError("must implement get_data()")
        # https://github.com/python/mypy/issues/5070
        # if False: yield is a mypy fix for
        # error: Return type "AsyncGenerator[Tuple[str, str], None]" of "get_data" incompatible with return type
        # "Coroutine[Any, Any, AsyncGenerator[Tuple[str, str], None]]" in supertype "BleCommunicationAsync"
        if False:
            yield 0



from unittest.mock import patch

from pytest import raises

from ruuvitag_sensor.adapters.dummy import BleCommunicationDummy
from ruuvitag_sensor.ruuvi import RuuviTagSensor
from ruuvitag_sensor.ruuvitag import RuuviTag


@patch("ruuvitag_sensor.ruuvi.ble", BleCommunicationDummy())
class TestRuuviTagSensor:
    def get_first_data(self, _mac, _bt_device=""):
        # https://ruu.vi/#AjwYAMFc
        data = "043E2A0201030157168974A5F41E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"
        return data[26:]

    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationDummy.get_first_data", get_first_data)
    def test_tag_update_is_valid(self):
        tag = RuuviTag("48:2C:6A:1E:59:3D")

        state = tag.state
        assert state == {}

        state = tag.update()
        assert state["temperature"] == 24
        assert state["pressure"] == 995
        assert state["humidity"] == 30

    def test_false_mac_raise_error(self):
        with raises(ValueError):
            RuuviTag("48:2C:6A:1E")

    def test_tag_correct_properties(self):
        org_mac = "AA:2C:6A:1E:59:3D"
        tag = RuuviTag(org_mac)

        mac = tag.mac
        state = tag.state
        assert mac == org_mac
        assert state == {}

    def get_data(self, _blacklist=None, _bt_device=""):
        if _blacklist is None:
            _blacklist = []

        tag_data = [
            ("AA:2C:6A:1E:59:3D", "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"),
            ("BB:2C:6A:1E:59:3D", "some other device"),
            ("CC:2C:6A:1E:59:3D", "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"),
            ("DD:2C:6A:1E:59:3D", "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"),
            ("EE:2C:6A:1E:59:3D", "1F0201060303AAFE1716AAFE10F9037275752E76692F23416A5558314D417730C3"),
            ("FF:2C:6A:1E:59:3D", "1902010415FF990403291A1ECE1E02DEF94202CA0B5300000000BB"),
            # The following entry is invalid, the data is corrupted
            ("00:2C:6A:1E:59:3D", "1902010415FF990403291A1ECE1E02DEF94202CA0B53BB"),
            ("11:2C:6A:1E:59:3D", "1F0201061BFF99040512FC5394C37C0004FFFC040CAC364200CDCBB8334C884FC4"),
            # The following entry is invalid, it has no MAC
            (None, "1F0201061BFF99040512FC5394C37C0004FFFC040CAC364200CDCBB8334C884FC4"),
            # The following is a ruuvitag with firmware 3.x, sending a non
            # measurement advertisement
            ("CE:D6:05:F5:17:AA", "1E11079ECADC240EE5A9E093F3A3B50100406E0B0952757576692031374141CB"),
            # The following is a ruuvitag with firmware 3.x, sending a
            # measurement advertisement again
            ("CE:D6:05:F5:17:AA", "1F0201061BFF99040511F83B83CC5DFFFCFFFC03DCA7161427D2CED605F517AACB"),
            # RuuviTag with firmware 3.3x sending RSSI
            ("D5:57:97:65:88:14", "1F0201061BFF99040517B24633FFFFFFFCFFD403E4AF56388D51D55797658814B8"),
        ]

        yield from tag_data

    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationDummy.get_data", get_data)
    def test_find_tags(self):
        tags = RuuviTagSensor.find_ruuvitags()
        assert len(tags) == 9

    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationDummy.get_data", get_data)
    def test_get_data_for_sensors(self):
        macs = ["CC:2C:6A:1E:59:3D", "DD:2C:6A:1E:59:3D", "EE:2C:6A:1E:59:3D", "D5:57:97:65:88:14"]
        data = RuuviTagSensor.get_data_for_sensors(macs, 4)
        assert len(data) == 4
        assert "CC:2C:6A:1E:59:3D" in data
        assert "DD:2C:6A:1E:59:3D" in data
        assert "D5:57:97:65:88:14" in data
        assert data["CC:2C:6A:1E:59:3D"]["temperature"] == 24.0
        assert data["EE:2C:6A:1E:59:3D"]["temperature"] == 25.12
        assert data["EE:2C:6A:1E:59:3D"]["identifier"] == "0"
        assert data["D5:57:97:65:88:14"]["rssi"] == -72

    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationDummy.get_data", get_data)
    def test_get_data(self):
        data = []
        RuuviTagSensor.get_data(data.append)
        assert len(data) == 9

    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationDummy.get_data", get_data)
    def test_get_data_with_macs(self):
        data = []
        macs = ["CC:2C:6A:1E:59:3D", "DD:2C:6A:1E:59:3D"]
        RuuviTagSensor.get_data(data.append, macs)
        assert len(data) == 2



import asyncio
from collections.abc import AsyncGenerator
from unittest.mock import patch

import pytest

from ruuvitag_sensor.adapters.dummy import BleCommunicationAsyncDummy, BleCommunicationDummy
from ruuvitag_sensor.ruuvi import RuuviTagSensor
from ruuvitag_sensor.ruuvi_types import MacAndRawData
from ruuvitag_sensor.ruuvitag import RuuviTagAsync


@pytest.mark.asyncio
class TestRuuviTagSensorAsync:
    async def _get_first_data(self, _mac, _bt_device):
        await asyncio.sleep(0)
        # https://ruu.vi/#AjwYAMFc
        data = "043E2A0201030157168974A5F41E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"
        return data[26:]

    async def _get_data(self, _blacklist=None, _bt_device="") -> AsyncGenerator[MacAndRawData, None]:
        if _blacklist is None:
            _blacklist = []

        tag_data = [
            ("EB:A5:D1:02:CE:68", "1c1bFF99040513844533c43dffe0ffd804189ff645fcffeba5d102ce68"),
            ("CD:D4:FA:52:7A:F2", "1c1bFF990405128a423bc45fffd8ff98040cafd6497a83cdd4fa527af2"),
            ("EC:4D:A7:95:08:6B", "1c1bFF9904050d3f5306c4df0024ffe404108f562787f1ec4da795086b"),
            ("CE:D6:05:F5:17:AA", "1c1bFF990405118f60bcc7f900880000fc0c813624238fc57d89639eb9a6"),
        ]

        for data in tag_data:
            yield data

    @patch("ruuvitag_sensor.ruuvi.ble", BleCommunicationAsyncDummy())
    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationAsyncDummy.get_data", _get_data)
    async def test_get_data_async(self):
        gener = RuuviTagSensor.get_data_async()
        data = [received async for received in gener]

        assert len(data) == 4

    @patch("ruuvitag_sensor.ruuvi.ble", BleCommunicationAsyncDummy())
    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationAsyncDummy.get_data", _get_data)
    async def test_get_data_async_with_macs(self):
        macs = ["EB:A5:D1:02:CE:68", "EC:4D:A7:95:08:6B"]
        gener = RuuviTagSensor.get_data_async(macs)
        data = [received async for received in gener]

        assert len(data) == 2

    @patch("ruuvitag_sensor.ruuvi.ble", BleCommunicationAsyncDummy())
    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationAsyncDummy.get_data", _get_data)
    async def test_find_ruuvitags_async_with_bleak(self):
        """Tests to see if MAC and the state of the sensors are returned for Bleak enabled request."""
        tags = await RuuviTagSensor.find_ruuvitags_async()

        assert len(tags) == 4

    @patch("ruuvitag_sensor.ruuvi.ble", BleCommunicationDummy())
    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationAsyncDummy.get_data", _get_data)
    async def test_find_ruuvitags_async_without_bleak(self):
        """Tests to see if exception is raised for non-Bleak enabled request.

        Async communication is supported only for BLE devices of subclass `BleCommunicationAsync`. This
        test forces a non-compatible BLE device type (BleCommunicationDummy subclass) to raise RuntimeError.
        """
        with pytest.raises(RuntimeError, match="Async BLE adapter required"):
            _ = await RuuviTagSensor.find_ruuvitags_async()

    @patch("ruuvitag_sensor.ruuvi.ble", BleCommunicationAsyncDummy())
    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationAsyncDummy.get_first_data", _get_first_data)
    async def test_tag_async_update_is_valid(self):
        tag = RuuviTagAsync("48:2C:6A:1E:59:3D")

        state = tag.state
        assert state == {}

        state = await tag.update()
        assert state["temperature"] == 24
        assert state["pressure"] == 995
        assert state["humidity"] == 30

    @patch("ruuvitag_sensor.ruuvi.ble", BleCommunicationAsyncDummy())
    @patch("ruuvitag_sensor.adapters.dummy.BleCommunicationAsyncDummy.get_data", _get_data)
    async def test_get_data_for_sensors(self):
        macs = ["EB:A5:D1:02:CE:68", "CD:D4:FA:52:7A:F2", "CE:D6:05:F5:17:AA"]
        data = await RuuviTagSensor.get_data_for_sensors_async(macs, 4)
        assert len(data) == 3
        assert "EB:A5:D1:02:CE:68" in data
        assert "CD:D4:FA:52:7A:F2" in data
        assert "CE:D6:05:F5:17:AA" in data
        assert data["EB:A5:D1:02:CE:68"]["temperature"] == 24.98
        assert data["CD:D4:FA:52:7A:F2"]["temperature"] == 23.73
        assert data["CE:D6:05:F5:17:AA"]["rssi"] == -90



from __future__ import annotations

import logging
import struct

from ruuvitag_sensor.ruuvi_types import ByteData, SensorDataE1

log = logging.getLogger(__name__)


class DfE1Decoder:
    """
    Decodes data from Ruuvi Air with Data Format E1
    Protocol specification:
    https://docs.ruuvi.com/communication/bluetooth-advertisements/data-format-e1.md
    """

    def _get_temperature(self, data: ByteData) -> float | None:
        """Return temperature in celsius"""
        if data[1] == -32768:
            return None
        return round(int(data[1]) * 0.005, 3)

    def _get_humidity(self, data: ByteData) -> float | None:
        """Return humidity %"""
        if data[2] == 65535:
            return None
        return round(int(data[2]) * 0.0025, 3)

    def _get_pressure(self, data: ByteData) -> float | None:
        """Return air pressure hPa"""
        if data[3] == 0xFFFF:
            return None
        return round((int(data[3]) + 50000) / 100, 2)

    def _get_pm1_ug_m3(self, data: ByteData) -> float | None:
        """Return PM 1.0, ug/m^3"""
        if data[4] == 0xFFFF:
            return None
        return round(int(data[4]) * 0.1, 1)

    def _get_pm25_ug_m3(self, data: ByteData) -> float | None:
        """Return PM 2.5, ug/m^3"""
        if data[5] == 0xFFFF:
            return None
        return round(int(data[5]) * 0.1, 1)

    def _get_pm4_ug_m3(self, data: ByteData) -> float | None:
        """Return PM 4.0, ug/m^3"""
        if data[6] == 0xFFFF:
            return None
        return round(int(data[6]) * 0.1, 1)

    def _get_pm10_ug_m3(self, data: ByteData) -> float | None:
        """Return PM 10.0, ug/m^3"""
        if data[7] == 0xFFFF:
            return None
        return round(int(data[7]) * 0.1, 1)

    def _get_co2_ppm(self, data: ByteData) -> int | None:
        """Return CO2 ppm"""
        if data[8] == 0xFFFF:
            return None
        return int(data[8])

    def _get_voc_index(self, data: ByteData) -> int | None:
        """Return VOC index (unitless, 9-bit)"""
        # VOC: bits [8:1] are in data[9], bit [0] (LSB) is FLAGS bit 6
        # As per spec: "9 bit unsigned, least significant bit in Flags byte"
        voc_high_bits = data[9]  # bits [8:1]
        voc_lsb = (data[14] >> 6) & 0x01  # bit [0]
        voc = (voc_high_bits << 1) | voc_lsb

        if voc == 511:  # 0x1FF
            return None

        return voc

    def _get_nox_index(self, data: ByteData) -> int | None:
        """Return NOx index (unitless, 9-bit)"""
        # NOx: bits [8:1] are in data[10], bit [0] (LSB) is FLAGS bit 7
        # As per spec: "9 bit unsigned, least significant bit in Flags byte"
        nox_high_bits = data[10]  # bits [8:1]
        nox_lsb = (data[14] >> 7) & 0x01  # bit [0]
        nox = (nox_high_bits << 1) | nox_lsb

        if nox == 511:  # 0x1FF
            return None

        return nox

    def _get_luminosity_lux(self, data: ByteData) -> float | None:
        """Return Luminosity Lux"""
        lumi_bytes = bytes(data[11])
        if lumi_bytes == b"\xff\xff\xff":
            return None
        lumi_val = int.from_bytes(lumi_bytes, byteorder="big")
        return round(lumi_val * 0.01, 2)

    def _get_measurementsequencenumber(self, data: ByteData) -> int | None:
        seq_bytes = bytes(data[13])
        if seq_bytes == b"\xff\xff\xff":
            return None
        return int.from_bytes(seq_bytes, byteorder="big")

    def _get_calibration_in_progress(self, data: ByteData) -> bool:
        # Bit 0 of flags indicates calibration status
        return bool(data[14] & 1)

    def _get_mac(self, data: ByteData) -> str:
        return ":".join(f"{b:02X}" for b in bytes(data[16]))

    def decode_data(self, data: str) -> SensorDataE1 | None:
        """
        Decode sensor data.

        Returns:
            dict: Sensor values
        """
        try:
            byte_data: ByteData = struct.unpack(">BhHHHHHHHBB3s3s3sB5s6s", bytearray.fromhex(data[:80]))
            return {
                "data_format": "E1",
                "humidity": self._get_humidity(byte_data),  # type: ignore
                "temperature": self._get_temperature(byte_data),  # type: ignore
                "pressure": self._get_pressure(byte_data),  # type: ignore
                "pm_1": self._get_pm1_ug_m3(byte_data),  # type: ignore
                "pm_2_5": self._get_pm25_ug_m3(byte_data),  # type: ignore
                "pm_4": self._get_pm4_ug_m3(byte_data),  # type: ignore
                "pm_10": self._get_pm10_ug_m3(byte_data),  # type: ignore
                "co2": self._get_co2_ppm(byte_data),  # type: ignore
                "voc": self._get_voc_index(byte_data),  # type: ignore
                "nox": self._get_nox_index(byte_data),  # type: ignore
                "luminosity": self._get_luminosity_lux(byte_data),  # type: ignore
                "measurement_sequence_number": self._get_measurementsequencenumber(byte_data),  # type: ignore
                "calibration_in_progress": self._get_calibration_in_progress(byte_data),
                "mac": self._get_mac(byte_data),
            }
        except Exception:
            log.exception("Value: %s not valid", data)
            return None



import asyncio
import random
from collections.abc import Awaitable, Callable
from dataclasses import dataclass

from bleak.backends.scanner import AdvertisementData, BLEDevice


@dataclass
class BLEDeviceDummy:
    address: str
    rssi: int


@dataclass
class AdvertisementDataDummy:
    manufacturer_data: dict


data = [
    ("CD:D4:FA:52:7A:F2", -89, b"\x05\x11\xd7E\xd6\xc8\xc2\xff\xfc\x00\x10\x04\x0c\xae\x16\x91{N\xcd\xd4\xfaRz\xf2"),
    ("FE:52:F7:B3:65:CC", -94, b"\x05\x08\xdas\xf0\xc9\r\xff\xc8\xff\xd8\x04\x00\x82\x16\xe0YN\xfeR\xf7\xb3e\xcc"),
    ("EC:4D:A7:95:08:6B", -84, b"\x05\x0c\x99U\x9d\xc9X\x00,\xff\xe8\x04\x10\x83\xd6}\x02F\xecM\xa7\x95\x08k"),
    ("E9:AC:CF:6E:C5:66", -88, b"\x05\x11ZH3\xc8\xc7\x00\x08\x00\x0c\x03\xdc\x83vJj\x94\xe9\xac\xcfn\xc5f"),
]


class DevBleakScanner:
    def __init__(self, callback, _):
        self.callback: Callable[[BLEDevice, AdvertisementData], Awaitable[None]] | None = callback  # type: ignore[annotation-unchecked]
        self.running: bool = False  # type: ignore[annotation-unchecked]

    async def start(self):
        self.running = True
        asyncio.create_task(self.run())

    async def stop(self):
        self.running = False

    async def run(self):
        while self.running:
            selected = random.choice(data)
            device = BLEDeviceDummy(address=selected[0], rssi=selected[1])
            advertisement_data = AdvertisementDataDummy(manufacturer_data={1177: selected[2]})
            if self.callback:
                await self.callback(device, advertisement_data)
            await asyncio.sleep(5)



from __future__ import annotations

import logging

from ruuvitag_sensor.ruuvi_types import SensorHistoryData

log = logging.getLogger(__name__)


class HistoryDecoder:
    """
    Decodes history data from RuuviTag
    Protocol specification:
    https://github.com/ruuvi/docs/blob/master/communication/bluetooth-connection/nordic-uart-service-nus/log-read.md

    Data format:
    - First byte: Command byte (0x3A)
    - Second byte: Packet type (0x30 = temperature, 0x31 = humidity, 0x32 = pressure)
    - Third byte: Header byte (skipped or error)
    - Next 4 bytes: Clock time (seconds since unix epoch)
    - Next 2 bytes: Reserved (always 0x00)
    - Next 2 bytes: Sensor data (uint16, big-endian)
        Temperature: 0.01°C units
        Humidity: 0.01% units
        Pressure: Raw value in hPa

    Special case:
    - End marker packet has command byte 0x3A followed by 0x3A
    """

    def _is_error_packet(self, data: list[str]) -> bool:
        """Check if this is an error packet"""
        return data[2] == "F0" and all(b == "ff" for b in data[3:])

    def _is_end_marker(self, data: list[str]) -> bool:
        """Check if this is an end marker packet"""
        return data[0] == "3a" and data[1] == "3a" and all(b == "ff" for b in data[3:])

    def _get_timestamp(self, data: list[str]) -> int:
        """Return timestamp"""
        # The timestamp is a 4-byte value after the header byte, in seconds since Unix epoch
        timestamp_bytes = bytes.fromhex("".join(data[3:7]))
        timestamp = int.from_bytes(timestamp_bytes, "big")
        return timestamp
        # return datetime.fromtimestamp(timestamp, tz=timezone.utc)

    def _get_temperature(self, data: list[str]) -> float | None:
        """Return temperature in celsius"""
        if data[1] != "30":  # '0' for temperature
            return None
        # Temperature is in 0.01°C units, big-endian
        temp_bytes = bytes.fromhex("".join(data[9:11]))
        temp_raw = int.from_bytes(temp_bytes, "big")
        return round(temp_raw * 0.01, 2)

    def _get_humidity(self, data: list[str]) -> float | None:
        """Return humidity %"""
        if data[1] != "31":  # '1' for humidity
            return None
        # Humidity is in 0.01% units, big-endian
        humidity_bytes = bytes.fromhex("".join(data[9:11]))
        humidity_raw = int.from_bytes(humidity_bytes, "big")
        return round(humidity_raw * 0.01, 2)

    def _get_pressure(self, data: list[str]) -> float | None:
        """Return air pressure hPa"""
        if data[1] != "32":  # '2' for pressure
            return None
        # Pressure is in hPa units, big-endian
        pressure_bytes = bytes.fromhex("".join(data[9:11]))
        pressure_raw = int.from_bytes(pressure_bytes, "big")
        return float(pressure_raw)

    def decode_data(self, data: bytearray) -> SensorHistoryData | None:  # noqa: PLR0911
        """
        Decode history data from RuuviTag.

        The data format follows the NUS log format.

        Args:
            data: Raw history data bytearray

        Returns:
            SensorDataHistory: Decoded sensor values with timestamp, or None if decoding fails
            Returns None for both invalid data and end marker packets
        """
        try:
            hex_values = [format(x, "02x") for x in data]

            if len(hex_values) != 11:
                if len(data) >= 5 and (data[0] == 0x3B or data[0] == 0xE1 or (len(data) >= 38 and data[4] == 0xE1)):
                    log.error(
                        "Received data appears to be from Ruuvi Air device. "
                        "Please use device_type='ruuvi_air' when calling get_history_async() or download_history()"
                    )
                else:
                    log.info("History data unexpected length: %d bytes (expected 11)", len(hex_values))
                return None

            if hex_values[0] != "3a":
                log.info("Invalid command byte: 0x%02X (expected 0x3A)", data[0])
                return None

            if self._is_error_packet(hex_values):
                log.info("Device reported error in log reading")
                return None

            if self._is_end_marker(hex_values):
                log.debug("End marker packet received")
                return None

            # Each packet type contains one measurement
            packet_type = hex_values[1]
            match packet_type:
                case "30":  # '0' temperature
                    return {
                        "temperature": self._get_temperature(hex_values),
                        "humidity": None,
                        "pressure": None,
                        "timestamp": self._get_timestamp(hex_values),
                    }
                case "31":  # '1' humidity
                    return {
                        "temperature": None,
                        "humidity": self._get_humidity(hex_values),
                        "pressure": None,
                        "timestamp": self._get_timestamp(hex_values),
                    }
                case "32":  # '2' pressure
                    return {
                        "temperature": None,
                        "humidity": None,
                        "pressure": self._get_pressure(hex_values),
                        "timestamp": self._get_timestamp(hex_values),
                    }
                case _:
                    log.info("Invalid packet type: %s - %s", packet_type, data)
                    return None

        except Exception:
            log.exception("Value not valid: %s", data)
            return None



"""
Decoder utilities and backward compatibility module.

This module contains decoder utility functions (get_decoder, parse_mac) and
re-exports all decoder classes from the decoders package for backward compatibility.
"""

from __future__ import annotations

import logging

# Re-export all decoder classes from the decoders package for backward compatibility.
# Existing code that imports from ruuvitag_sensor.decoder will continue to work.
# New code should import directly from ruuvitag_sensor.decoders for better clarity.
from ruuvitag_sensor.decoders import (
    AirHistoryDecoder,
    Df3Decoder,
    Df5Decoder,
    Df6Decoder,
    DfE1Decoder,
    HistoryDecoder,
    UrlDecoder,
)

log = logging.getLogger(__name__)


__all__ = [
    # Decoder classes (re-exported from ruuvitag_sensor.decoders)
    "AirHistoryDecoder",
    "Df3Decoder",
    "Df5Decoder",
    "Df6Decoder",
    "DfE1Decoder",
    "HistoryDecoder",
    "UrlDecoder",
    # Utility functions (defined in this module)
    "get_decoder",
    "parse_mac",
]


def get_decoder(data_format: int | str) -> UrlDecoder | Df3Decoder | Df5Decoder | Df6Decoder | DfE1Decoder:
    """
    Get correct decoder for Data Format.

    Args:
        data_format: The data format number (2, 3, 4, 5, 6, or "E1")

    Returns:
        object: Data decoder instance

    Raises:
        ValueError: If data_format is not a recognized format
    """
    # Data formats are ordered by priority, so the most common formats are at the top.
    match data_format:
        case 5:
            # https://docs.ruuvi.com/communication/bluetooth-advertisements/data-format-5-rawv2
            return Df5Decoder()
        case 6:
            # https://docs.ruuvi.com/communication/bluetooth-advertisements/data-format-6
            return Df6Decoder()
        case "E1":
            # https://docs.ruuvi.com/communication/bluetooth-advertisements/data-format-e1.md
            return DfE1Decoder()
        case 2:
            log.warning("DATA TYPE 2 IS OBSOLETE. UPDATE YOUR TAG")
            # https://github.com/ruuvi/ruuvi-sensor-protocols/blob/master/dataformat_04.md
            return UrlDecoder()
        case 3:
            log.warning("DATA TYPE 3 IS DEPRECATED - UPDATE YOUR TAG")
            # https://github.com/ruuvi/ruuvi-sensor-protocols/blob/master/dataformat_03.md
            return Df3Decoder()
        case 4:
            log.warning("DATA TYPE 4 IS OBSOLETE. UPDATE YOUR TAG")
            # https://github.com/ruuvi/ruuvi-sensor-protocols/blob/master/dataformat_04.md
            return UrlDecoder()
        case _:
            # This should never happen in normal operation since DataFormats.convert_data()
            # already validates and identifies the data format. If we reach here, it indicates
            # a programming error (e.g., convert_data was bypassed or returned an unhandled format).
            raise ValueError(f"Unknown data format: {data_format}")


def parse_mac(data_format: int | str, payload_mac: str) -> str:
    """
    Parse MAC address from payload data.

    Data format 5 payload contains MAC-address in format e.g. e62eb92e73e5

    Args:
        data_format: The data format number
        payload_mac: MAC address string from payload

    Returns:
        string: MAC separated and in upper case e.g. E6:2E:B9:2E:73:E5
    """
    match data_format:
        case 5:
            return ":".join(payload_mac[i : i + 2] for i in range(0, 12, 2)).upper()
        case _:
            return payload_mac



from __future__ import annotations

import logging

from ruuvitag_sensor.ruuvi_types import SensorAirHistoryData

log = logging.getLogger(__name__)


class AirHistoryDecoder:
    """
    Decodes history data from Ruuvi Air
    Protocol specification:
    https://docs.ruuvi.com/communication/bluetooth-connection/nordic-uart-service-nus/log-read-air.md

    Multi-record response packet format:
    - Byte 0: Destination (source index from request)
    - Byte 1: Source (0x3B = Air Quality endpoint)
    - Byte 2: Operation (0x20 = Multi-record log write)
    - Byte 3: Num records (number of records in this packet)
    - Byte 4: Record length (38 bytes)
    - Bytes 5+: Packed record data

    Each record (38 bytes):
    - Bytes 0-3: Timestamp (uint32_t BE, Unix timestamp in seconds)
    - Byte 4: Data format (0xE1)
    - Bytes 5-6: Temperature (int16_t BE, x 200)
    - Bytes 7-8: Humidity (uint16_t BE, x 400)
    - Bytes 9-10: Pressure (uint16_t BE, (Pressure - 50000) Pa)
    - Bytes 11-12: PM1.0 (uint16_t BE, x 10)
    - Bytes 13-14: PM2.5 (uint16_t BE, x 10)
    - Bytes 15-16: PM4.0 (uint16_t BE, x 10)
    - Bytes 17-18: PM10.0 (uint16_t BE, x 10)
    - Bytes 19-20: CO₂ (uint16_t BE, ppm)
    - Byte 21: VOC (uint8_t, bit 9 in flags)
    - Byte 22: NOx (uint8_t, bit 9 in flags)
    - Bytes 23-28: Reserved
    - Bytes 29-31: Sequence counter (uint24_t BE)
    - Byte 32: Flags (extended bits for 9-bit values)
    - Bytes 33-37: Reserved
    """

    def _is_end_marker(self, data: bytearray) -> bool:
        """Check if this is an end marker packet (num_records = 0)."""
        if len(data) < 5:
            return False
        # End marker: destination=0x3B, source=0x3B, operation=0x20, num_records=0, record_length=38
        return data[0] == 0x3B and data[1] == 0x3B and data[2] == 0x20 and data[3] == 0x00 and data[4] == 0x26

    def _get_timestamp(self, record_data: bytearray) -> int:
        """Extract timestamp from record (bytes 0-3, big-endian)."""
        return int.from_bytes(record_data[0:4], byteorder="big")

    def _get_temperature(self, record_data: bytearray) -> float | None:
        """Extract temperature from record (bytes 5-6, int16_t BE, x 200)."""
        temp_raw = int.from_bytes(record_data[5:7], byteorder="big", signed=False)
        if temp_raw == 0x8000:
            return None
        if temp_raw >= 0x8000:
            temp_raw = temp_raw - 0x10000
        return round(temp_raw / 200.0, 2)

    def _get_humidity(self, record_data: bytearray) -> float | None:
        """Extract humidity from record (bytes 7-8, uint16_t BE, x 400)."""
        humidity_raw = int.from_bytes(record_data[7:9], byteorder="big")
        return None if humidity_raw == 0xFFFF else round(humidity_raw / 400.0, 3)

    def _get_pressure(self, record_data: bytearray) -> float | None:
        """Extract pressure from record (bytes 9-10, uint16_t BE)."""
        pressure_raw = int.from_bytes(record_data[9:11], byteorder="big")
        return None if pressure_raw == 0xFFFF else round((pressure_raw + 50000) / 100.0, 2)

    def _get_pm_value(self, record_data: bytearray, offset: int) -> float | None:
        """Extract PM value from record (uint16_t BE, x 10)."""
        pm_raw = int.from_bytes(record_data[offset : offset + 2], byteorder="big")
        return None if pm_raw == 0xFFFF else round(pm_raw / 10.0, 1)

    def _get_co2(self, record_data: bytearray) -> int | None:
        """Extract CO2 from record (bytes 19-20, uint16_t BE)."""
        co2_raw = int.from_bytes(record_data[19:21], byteorder="big")
        return None if co2_raw == 0xFFFF else co2_raw

    def _get_9bit_value(self, record_data: bytearray, byte_offset: int, flag_bit: int) -> int | None:
        """Extract 9-bit value from record (8 bits in byte + 1 bit in flags)."""
        flags = record_data[32] if len(record_data) > 32 else 0
        value_byte = record_data[byte_offset]
        value_bit9 = (flags >> flag_bit) & 0x01
        value = value_byte | (value_bit9 << 8)
        return None if value == 0x1FF else value

    def _get_sequence(self, record_data: bytearray) -> int | None:
        """Extract sequence counter from record (bytes 29-31, uint24_t BE)."""
        seq_raw = int.from_bytes(record_data[29:32], byteorder="big")
        return None if seq_raw == 0xFFFFFF else seq_raw

    def _decode_record(self, record_data: bytearray) -> SensorAirHistoryData | None:
        """
        Decode a single 38-byte record.

        Args:
            record_data: 38-byte record data

        Returns:
            SensorAirHistoryData or None if invalid
        """
        if len(record_data) < 38:
            log.debug("Record too short: %d bytes", len(record_data))
            return None

        try:
            if record_data[4] != 0xE1:
                log.debug("Invalid data format: 0x%02X (expected 0xE1)", record_data[4])
                return None

            return {
                "timestamp": self._get_timestamp(record_data),
                "temperature": self._get_temperature(record_data),
                "humidity": self._get_humidity(record_data),
                "pressure": self._get_pressure(record_data),
                "pm_1": self._get_pm_value(record_data, 11),
                "pm_2_5": self._get_pm_value(record_data, 13),
                "pm_4": self._get_pm_value(record_data, 15),
                "pm_10": self._get_pm_value(record_data, 17),
                "co2": self._get_co2(record_data),
                "voc": self._get_9bit_value(record_data, 21, 6),
                "nox": self._get_9bit_value(record_data, 22, 7),
                "measurement_sequence_number": self._get_sequence(record_data),
            }

        except Exception:
            log.exception("Failed to decode record")
            return None

    def decode_data(self, data: bytearray) -> list[SensorAirHistoryData]:
        """
        Decode a multi-record response packet.

        Args:
            data: Raw packet data from BLE notification

        Returns:
            List of decoded records (empty list if end marker or invalid packet)
        """
        if len(data) < 5:
            log.debug("Packet too short: %d bytes", len(data))
            return []

        if self._is_end_marker(data):
            log.debug("End marker received")
            return []

        if data[0] != 0x3B or data[1] != 0x3B or data[2] != 0x20:
            log.debug("Invalid packet header: 0x%02X 0x%02X 0x%02X", data[0], data[1], data[2])
            return []

        num_records = data[3]
        record_length = data[4]

        if record_length != 38:
            log.debug("Unexpected record length: %d (expected 38)", record_length)
            return []

        if num_records == 0:
            log.debug("No records in packet")
            return []

        records = []
        header_size = 5
        for i in range(num_records):
            record_start = header_size + (i * record_length)
            record_end = record_start + record_length
            if record_end > len(data):
                log.debug("Not enough data for record %d", i)
                break
            record_data = data[record_start:record_end]
            if decoded := self._decode_record(record_data):
                records.append(decoded)

        return records



import asyncio
import time
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime, timezone
from multiprocessing import Manager
from multiprocessing.managers import DictProxy
from queue import Queue
from threading import Thread

from reactivex import Subject

from ruuvitag_sensor.adapters import is_async_adapter
from ruuvitag_sensor.ruuvi import RunFlag, RuuviTagSensor, ble


async def _run_get_data_background_async(macs: list[str], queue: Queue, shared_data: DictProxy, bt_device: str):
    """
    Async background process function for RuuviTag Sensors
    """
    data_iter = RuuviTagSensor.get_data_async(macs, bt_device)
    try:
        async for data in data_iter:
            if not shared_data["run_flag"]:
                break

            data[1]["time"] = datetime.now(timezone.utc).isoformat()  # type: ignore
            queue.put(data)
    finally:
        await data_iter.aclose()


def _run_get_data_background(macs: list[str], queue: Queue, shared_data: DictProxy, bt_device: str):
    """
    Background process function for RuuviTag Sensors
    """

    run_flag = RunFlag()

    def add_data(data):
        if not shared_data["run_flag"]:
            run_flag.running = False

        data[1]["time"] = datetime.now(timezone.utc).isoformat()
        queue.put(data)

    RuuviTagSensor.get_data(add_data, macs, run_flag, bt_device)


class RuuviTagReactive:
    """
    Reactive wrapper and background process for RuuviTagSensor get_data
    """

    @staticmethod
    def _data_update(subjects: list[Subject], queue: Queue, run_flag: RunFlag):
        """
        Get data from background process and notify all subscribed observers with the new data
        """
        while run_flag.running:
            while not queue.empty():
                data = queue.get()
                for subject in [s for s in subjects if not s.is_disposed]:
                    subject.on_next(data)
            time.sleep(0.1)

    def __init__(self, macs: list[str] | None = None, bt_device: str = ""):
        """
        Start background process for get_data and async task for notifying all subscribed observers

        Args:
            macs (list): MAC addresses
            bt_device (string): Bluetooth device id
        """
        if macs is None:
            macs = []

        self._run_flag = RunFlag()
        self._subjects: list[Subject] = []

        m = Manager()
        q = m.Queue()

        # Use Manager dict to share data between processes
        self._shared_data = m.dict()
        self._shared_data["run_flag"] = True

        # Start data updater

        notify_thread = Thread(target=RuuviTagReactive._data_update, args=(self._subjects, q, self._run_flag))
        notify_thread.start()

        # Start background process

        if is_async_adapter(ble):
            loop = asyncio.get_running_loop()
            loop.create_task(_run_get_data_background_async(macs, q, self._shared_data, bt_device))
        else:
            executor = ProcessPoolExecutor(1)
            executor.submit(_run_get_data_background, macs, q, self._shared_data, bt_device)

    def get_subject(self) -> Subject:
        """
        Returns:
            subject : Reactive Extension Subject
        """

        if not self._run_flag.running:
            raise Exception("RuuviTagReactive stopped")

        subject: Subject = Subject()
        self._subjects.append(subject)
        return subject

    def stop(self):
        """
        Stop get_data
        """

        self._run_flag.running = False
        self._shared_data["run_flag"] = False

        for s in self._subjects:
            s.dispose()



from __future__ import annotations

import logging
import math
import struct

from ruuvitag_sensor.ruuvi_types import ByteData, SensorData6

log = logging.getLogger(__name__)


class Df6Decoder:
    """
    Decodes data from Ruuvi Air with Data Format 6
    Protocol specification:
    https://docs.ruuvi.com/communication/bluetooth-advertisements/data-format-6

    Data Format 6 is used by Ruuvi Air for air quality monitoring.
    It includes CO2, PM2.5, VOC, NOx, luminosity along with traditional sensors.
    """

    def _get_temperature(self, data: ByteData) -> float | None:
        """Return temperature in celsius"""
        if data[1] == -32768:
            return None

        return round(data[1] * 0.005, 3)

    def _get_humidity(self, data: ByteData) -> float | None:
        """Return humidity %"""
        if data[2] == 65535:
            return None

        return round(data[2] * 0.0025, 3)

    def _get_pressure(self, data: ByteData) -> float | None:
        """Return air pressure hPa"""
        if data[3] == 65535:
            return None

        return round((data[3] + 50000) / 100, 2)

    def _get_pm_2_5(self, data: ByteData) -> float | None:
        """Return PM 2.5 in ug/m³"""
        if data[4] == 65535:
            return None

        return round(data[4] * 0.1, 1)

    def _get_co2(self, data: ByteData) -> int | None:
        """Return CO2 concentration in ppm"""
        if data[5] == 65535:
            return None

        return data[5]

    def _get_voc(self, data: ByteData) -> int | None:
        """Return VOC index (unitless, 9-bit)"""
        # VOC: bits [8:1] are in data[6], bit [0] (LSB) is FLAGS bit 6
        # As per spec: "9 bit unsigned, least significant bit in Flags byte"
        voc_high_bits = data[6]  # bits [8:1]
        voc_lsb = (data[11] >> 6) & 0x01  # bit [0]
        voc = (voc_high_bits << 1) | voc_lsb

        if voc == 511:  # 0x1FF
            return None

        return voc

    def _get_nox(self, data: ByteData) -> int | None:
        """Return NOx index (unitless, 9-bit)"""
        # NOx: bits [8:1] are in data[7], bit [0] (LSB) is FLAGS bit 7
        # As per spec: "9 bit unsigned, least significant bit in Flags byte"
        nox_high_bits = data[7]  # bits [8:1]
        nox_lsb = (data[11] >> 7) & 0x01  # bit [0]
        nox = (nox_high_bits << 1) | nox_lsb

        if nox == 511:  # 0x1FF
            return None

        return nox

    def _get_luminosity(self, data: ByteData) -> float | None:
        """Return luminosity in lux (logarithmic scale)"""
        code = data[8]

        if code == 255:
            return None

        if code == 0:
            return 0.0

        # Decode logarithmic value
        MAX_VALUE = 65535
        MAX_CODE = 254
        DELTA = math.log(MAX_VALUE + 1) / MAX_CODE
        value = math.exp(code * DELTA) - 1

        return round(value, 2)

    def _get_measurement_sequence_number(self, data: ByteData) -> int:
        """Return measurement sequence number"""
        return data[10]

    def _get_calibration_in_progress(self, data: ByteData) -> bool:
        """Return calibration status from flags byte"""
        # Bit 0 of flags byte indicates calibration status
        return bool(data[11] & 0x01)

    def _get_mac(self, data: ByteData) -> str:
        """Return MAC address (last 3 bytes)"""
        return "".join(f"{x:02x}" for x in data[12:15])

    def decode_data(self, data: str) -> SensorData6 | None:
        """
        Decode sensor data.

        Returns:
            dict: Sensor values
        """
        try:
            # Data Format 6 structure (20 bytes):
            # 0: data_format (uint8)
            # 1-2: temperature (int16)
            # 3-4: humidity (uint16)
            # 5-6: pressure (uint16)
            # 7-8: pm_2_5 (uint16)
            # 9-10: co2 (uint16)
            # 11: voc_low (uint8)
            # 12: nox_low (uint8)
            # 13: luminosity (uint8)
            # 14: reserved (uint8)
            # 15: measurement_sequence (uint8)
            # 16: flags (uint8)
            # 17-19: mac (3 bytes)
            byte_data: ByteData = struct.unpack(">BhHHHHBBBBBBBBB", bytearray.fromhex(data[:40]))

            return {
                "data_format": 6,
                "temperature": self._get_temperature(byte_data),  # type: ignore
                "humidity": self._get_humidity(byte_data),  # type: ignore
                "pressure": self._get_pressure(byte_data),  # type: ignore
                "pm_2_5": self._get_pm_2_5(byte_data),  # type: ignore
                "co2": self._get_co2(byte_data),  # type: ignore
                "voc": self._get_voc(byte_data),  # type: ignore
                "nox": self._get_nox(byte_data),  # type: ignore
                "luminosity": self._get_luminosity(byte_data),  # type: ignore
                "measurement_sequence_number": self._get_measurement_sequence_number(byte_data),
                "calibration_in_progress": self._get_calibration_in_progress(byte_data),
                "mac": self._get_mac(byte_data),
            }
        except Exception:
            log.exception("Value: %s not valid", data)
            return None



import logging
import sys
import time
from collections.abc import Generator
from multiprocessing import Manager, Process
from queue import Queue

from bleson import Observer, get_provider

from ruuvitag_sensor.adapters import BleCommunication
from ruuvitag_sensor.adapters.utils import rssi_to_hex
from ruuvitag_sensor.ruuvi_types import MacAndRawData, RawData

log = logging.getLogger(__name__)


class BleCommunicationBleson(BleCommunication):
    """Bluetooth LE communication with Bleson"""

    @staticmethod
    def _run_get_data_background(queue, shared_data, bt_device):
        (observer, q) = BleCommunicationBleson.start(bt_device)

        for advertisement in BleCommunicationBleson.get_lines(q):
            log.debug("Data: %s", advertisement)
            if shared_data["stop"]:
                break
            try:
                # macOS doesn't return address on advertised package
                mac = advertisement.address.address if advertisement.address is not None else None
                if mac and mac in shared_data["blacklist"]:
                    log.debug("MAC blacklised: %s", mac)
                    continue
                if advertisement.mfg_data is None:
                    continue
                # Linux returns bytearray for mfg_data, but macOS returns _NSInlineData
                # which casts to byte array. We need to explicitly cast it to use hex
                data = (
                    bytearray(advertisement.mfg_data) if sys.platform.startswith("darwin") else advertisement.mfg_data
                )
                # Bleson returns data in a different format than the nix_hci
                # adapter. Since the rest of the processing pipeline is
                # somewhat reliant on the additional data, add to the
                # beginning of the actual data:
                #
                # - An FF type marker
                # - A length marker, covering the vendor specific data
                # - Another length marker, covering the length-marked
                #   vendor data.
                #
                # Thus extended, the result can be parsed by the rest of
                # the pipeline.
                #
                # TODO: This is kinda awkward, and should be handled better.
                data = f"FF{data.hex()}"
                data = f"{(len(data) >> 1):02x}{data}"
                data = f"{(len(data) >> 1):02x}{data}"

                # Add RSSI to encoded data as hex. All adapters use a common decoder.
                data += rssi_to_hex(advertisement.rssi)
                queue.put((mac, data.upper()))
            except GeneratorExit:
                break
            except Exception:
                log.exception("Error in advertisement handling")
                continue

        BleCommunicationBleson.stop(observer)

    @staticmethod
    def start(bt_device=""):
        """
        Attributes:
           device (string): BLE device (default 0)
        """

        # Old communication used hci0 etc.
        bt_device = 0 if not bt_device else bt_device.replace("hci", "")

        log.info("Start receiving broadcasts (device %s)", bt_device)

        q = Queue()

        adapter = get_provider().get_adapter(int(bt_device))
        observer = Observer(adapter)
        observer.on_advertising_data = q.put
        observer.start()

        return (observer, q)

    @staticmethod
    def stop(observer):
        observer.stop()

    @staticmethod
    def get_lines(queue):
        try:
            while True:
                next_item = queue.get(True, None)
                yield next_item
        except KeyboardInterrupt:
            return
        except Exception as ex:
            log.info(ex)
            return

    @staticmethod
    def get_data(blacklist: list[str] | None = None, bt_device: str = "") -> Generator[MacAndRawData, None, None]:
        m = Manager()
        q = m.Queue()

        # Use Manager dict to share data between processes
        shared_data = m.dict()
        shared_data["blacklist"] = blacklist or []
        shared_data["stop"] = False

        # Start background process
        proc = Process(target=BleCommunicationBleson._run_get_data_background, args=[q, shared_data, bt_device])
        proc.start()

        try:
            while True:
                while not q.empty():
                    data = q.get()
                    yield data
                time.sleep(0.1)
        except GeneratorExit:
            pass

        shared_data["stop"] = True
        proc.join()

    @staticmethod
    def get_first_data(mac: str, bt_device: str = "") -> RawData:
        data = None
        data_iter = BleCommunicationBleson.get_data([], bt_device)
        for d in data_iter:
            if mac == d[0]:
                log.info("Data found")
                data_iter.close()
                data = d[1]
                break

        return data or ""



from collections.abc import AsyncGenerator, Generator

from ruuvitag_sensor.adapters import BleCommunication, BleCommunicationAsync
from ruuvitag_sensor.ruuvi_types import MacAndRawData, RawData


class BleCommunicationDummy(BleCommunication):
    """TODO: Find some working BLE implementation for Windows and OSX"""

    @staticmethod
    def get_first_data(_mac: str, _bt_device: str = "") -> RawData:
        return "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"

    @staticmethod
    def get_data(_blacklist: list[str] | None = None, _bt_device: str = "") -> Generator[MacAndRawData, None, None]:
        if _blacklist is None:
            _blacklist = []

        dummy_data = [
            ("DU:MM:YD:AT:A9:3D", "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"),
            ("NO:TS:UP:PO:RT:ED", "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"),
        ]

        yield from dummy_data


class BleCommunicationAsyncDummy(BleCommunicationAsync):
    """TODO: Find some working BLE implementation for Windows and OSX"""

    @staticmethod
    async def get_first_data(_mac: str, _bt_device: str = "") -> RawData:
        return "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"

    @staticmethod
    async def get_data(
        _blacklist: list[str] | None = None, _bt_device: str = ""
    ) -> AsyncGenerator[MacAndRawData, None]:
        if _blacklist is None:
            _blacklist = []

        dummy_data = [
            ("DU:MM:YD:AT:A9:3D", "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"),
            ("NO:TS:UP:PO:RT:ED", "1E0201060303AAFE1616AAFE10EE037275752E76692F23416A7759414D4663CD"),
        ]

        for data in dummy_data:
            yield data



"""
Decoder modules for different RuuviTag and Ruuvi Air data formats.

This package contains individual decoder modules for each data format:
- url_decoder: DF2/DF4 (URL-encoded formats, deprecated)
- df3_decoder: DF3 (raw sensor data, deprecated)
- df5_decoder: DF5 (extended raw data, primary RuuviTag format)
- df6_decoder: DF6 (Ruuvi Air quality data)
- dfe1_decoder: DFE1 (extended Ruuvi Air quality data)
- history_decoder: RuuviTag history data decoder
- air_history_decoder: Ruuvi Air history data decoder
"""

from ruuvitag_sensor.decoders.air_history_decoder import AirHistoryDecoder
from ruuvitag_sensor.decoders.df3_decoder import Df3Decoder
from ruuvitag_sensor.decoders.df5_decoder import Df5Decoder
from ruuvitag_sensor.decoders.df6_decoder import Df6Decoder
from ruuvitag_sensor.decoders.dfe1_decoder import DfE1Decoder
from ruuvitag_sensor.decoders.history_decoder import HistoryDecoder
from ruuvitag_sensor.decoders.url_decoder import UrlDecoder

__all__ = [
    "AirHistoryDecoder",
    "Df3Decoder",
    "Df5Decoder",
    "Df6Decoder",
    "DfE1Decoder",
    "HistoryDecoder",
    "UrlDecoder",
]



import argparse
import asyncio
import logging
import sys

import ruuvitag_sensor
from ruuvitag_sensor.adapters import is_async_adapter
from ruuvitag_sensor.log import log
from ruuvitag_sensor.ruuvi import RuuviTagSensor

ruuvitag_sensor.log.enable_console()


def my_excepthook(exctype, value, traceback):
    sys.__excepthook__(exctype, value, traceback)

    if not issubclass(exctype, KeyboardInterrupt):
        log.critical(value)


sys.excepthook = my_excepthook


async def _async_main_handle(arguments: argparse.Namespace):
    if arguments.mac_address:
        data = await RuuviTagSensor.get_data_for_sensors_async(
            macs=[arguments.mac_address], bt_device=arguments.bt_device
        )
        log.info(data)
    elif arguments.find_action:
        await RuuviTagSensor.find_ruuvitags_async(arguments.bt_device)
    elif arguments.latest_action:
        data = await RuuviTagSensor.get_data_for_sensors_async(bt_device=arguments.bt_device)
        log.info(data)
    elif arguments.stream_action:
        async for mac, sensor_data in RuuviTagSensor.get_data_async(bt_device=arguments.bt_device):
            log.info("%s - %s", mac, sensor_data)


def _sync_main_handle(arguments: argparse.Namespace):
    if arguments.mac_address:
        data = RuuviTagSensor.get_data_for_sensors(macs=[arguments.mac_address], bt_device=arguments.bt_device)
        log.info(data)
    elif arguments.find_action:
        RuuviTagSensor.find_ruuvitags(arguments.bt_device)
    elif arguments.latest_action:
        data = RuuviTagSensor.get_data_for_sensors(bt_device=arguments.bt_device)
        log.info(data)
    elif arguments.stream_action:
        RuuviTagSensor.get_data(lambda x: log.info("%s - %s", x[0], x[1]), bt_device=arguments.bt_device)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-g", "--get", dest="mac_address", help="Get data")
    parser.add_argument("-d", "--device", dest="bt_device", help="Set Bluetooth device id (default hci0)")
    parser.add_argument("-f", "--find", action="store_true", dest="find_action", help="Find broadcasting RuuviTags")
    parser.add_argument(
        "-l", "--latest", action="store_true", dest="latest_action", help="Get latest data for found RuuviTags"
    )
    parser.add_argument(
        "-s", "--stream", action="store_true", dest="stream_action", help="Stream broadcasts from all RuuviTags"
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {ruuvitag_sensor.__version__}")
    parser.add_argument("--debug", action="store_true", dest="debug_action", help="Enable debug logging")
    args = parser.parse_args()

    if args.debug_action:
        log.setLevel(logging.DEBUG)
        for handler in log.handlers:
            handler.setLevel(logging.DEBUG)

    if not args.mac_address and not args.find_action and not args.latest_action and not args.stream_action:
        parser.print_usage()
        sys.exit(0)

    if is_async_adapter(ruuvitag_sensor.ruuvi.ble):
        asyncio.run(_async_main_handle(args))
    else:
        _sync_main_handle(args)



import asyncio
import logging
import os
import re
import sys
from collections.abc import AsyncGenerator, Callable
from datetime import datetime, timezone
from enum import Enum

from bleak import BleakClient, BleakGATTCharacteristic, BleakScanner
from bleak.backends.scanner import AdvertisementData, AdvertisementDataCallback, BLEDevice

from ruuvitag_sensor.adapters import BleCommunicationAsync
from ruuvitag_sensor.adapters.utils import rssi_to_hex
from ruuvitag_sensor.ruuvi_types import MacAndRawData, RawData

MAC_REGEX = "[0-9a-f]{2}([:])[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$"
RUUVI_HISTORY_SERVICE_UUID = "6E400001-B5A3-F393-E0A9-E50E24DCCA9E"
RUUVI_HISTORY_RX_CHAR_UUID = "6E400002-B5A3-F393-E0A9-E50E24DCCA9E"  # Write
RUUVI_HISTORY_TX_CHAR_UUID = "6E400003-B5A3-F393-E0A9-E50E24DCCA9E"  # Read and notify


def _get_scanner(detection_callback: AdvertisementDataCallback, bt_device: str = ""):
    # NOTE: On Linux - bleak.exc.BleakError: passive scanning mode requires bluez or_patterns
    # NOTE: On macOS - bleak.exc.BleakError: macOS does not support passive scanning
    scanning_mode = "passive" if sys.platform.startswith("win") else "active"

    if "bleak_dev" in os.environ.get("RUUVI_BLE_ADAPTER", "").lower():
        from ruuvitag_sensor.adapters.development.dev_bleak_scanner import DevBleakScanner  # noqa: PLC0415

        return DevBleakScanner(detection_callback, scanning_mode)

    if bt_device:
        return BleakScanner(
            detection_callback=detection_callback,
            scanning_mode=scanning_mode,  # type: ignore[arg-type]
            adapter=bt_device,
        )

    return BleakScanner(detection_callback=detection_callback, scanning_mode=scanning_mode)  # type: ignore[arg-type]


queue = asyncio.Queue[tuple[str, str]]()

log = logging.getLogger(__name__)


class HistoryNotificationAction(str, Enum):
    """Action types for history notification processing."""

    IGNORE = "ignore"  # Heartbeat data, should be ignored
    END = "end"  # End marker received
    ERROR = "error"  # Error packet received
    DATA = "data"  # Normal history data


class BleCommunicationBleak(BleCommunicationAsync):
    @staticmethod
    def _parse_data(data: bytes) -> str:
        # Bleak returns data in a different format than the nix_hci
        # adapter. Since the rest of the processing pipeline is
        # somewhat reliant on the additional data, add to the
        # beginning of the actual data:
        #
        # - An FF type marker with 9904 (Ruuvi manufacturer identifier)
        # - A length marker, covering the vendor specific data
        # - Another length marker, covering the length-marked
        #   vendor data.
        #
        # Thus extended, the result can be parsed by the rest of
        # the pipeline.
        #
        # TODO: This is kinda awkward, and should be handled better.
        formatted = f"FF9904{data.hex()}"
        formatted = f"{(len(formatted) >> 1):02x}{formatted}"
        formatted = f"{(len(formatted) >> 1):02x}{formatted}"
        return formatted

    @staticmethod
    async def get_data(blacklist: list[str] | None = None, bt_device: str = "") -> AsyncGenerator[MacAndRawData, None]:
        async def detection_callback(device: BLEDevice, advertisement_data: AdvertisementData):
            # On macOS device address is not a MAC address, but a system specific ID
            # https://github.com/hbldh/bleak/issues/140
            mac: str = device.address if re.match(MAC_REGEX, device.address.lower()) else ""
            if blacklist and mac in blacklist:
                log.debug("MAC blacklised: %s", mac)
                return

            # TODO: Do all RuuviTags have data in 1177?
            if 1177 not in advertisement_data.manufacturer_data:
                return

            log.debug("Received data: %s", advertisement_data)

            data = BleCommunicationBleak._parse_data(advertisement_data.manufacturer_data[1177])

            # Add RSSI to encoded data as hex. All adapters use a common decoder.
            data += rssi_to_hex(advertisement_data.rssi)
            await queue.put((mac, data))

        scanner = _get_scanner(detection_callback, bt_device)
        await scanner.start()

        log.debug("Bleak scanner started")

        try:
            while True:
                next_item: tuple[str, str] = await queue.get()
                yield next_item
        except KeyboardInterrupt:
            pass
        except GeneratorExit:
            pass
        except Exception as ex:
            log.info(ex)

        await scanner.stop()

        log.debug("Bleak scanner stopped")

    @staticmethod
    async def get_first_data(mac: str, bt_device: str = "") -> RawData:
        """
        NOTE: get_first_data does not work on macOS.

        macOS doesn't return MAC address, as it uses system specific IDs
        """
        data = None
        data_iter = BleCommunicationBleak.get_data([], bt_device)
        async for d in data_iter:
            if mac == d[0]:
                log.info("Data found")
                data = d[1]
                await data_iter.aclose()

        return data or ""

    async def get_history_data(
        self,
        mac: str,
        start_time: datetime | None = None,
        max_items: int | None = None,
        device_type: str = "ruuvitag",
    ) -> AsyncGenerator[bytearray, None]:
        """
        Get history data from a RuuviTag or Ruuvi Air using GATT connection.

        Args:
            mac (str): MAC address or UUID of the device
            start_time (datetime, optional): Start time for history data. Time should be in UTC.
            max_items (int, optional): Maximum number of history entries to fetch
            device_type (str): Device type - "ruuvi_air" for Ruuvi Air,
                "ruuvitag" for RuuviTag (default: "ruuvitag")

        Yields:
            bytearray: Raw history data entries (packets for Ruuvi Air, single records for RuuviTag)

        Raises:
            RuntimeError: If connection fails or required services not found
        """
        client = None
        try:
            log.debug("Connecting to device %s", mac)
            client = await self._connect_gatt(mac)
            log.debug("Connected to device %s", mac)

            is_ruuvi_air = device_type == "ruuvi_air"

            # Try to negotiate larger MTU for Ruuvi Air (recommended: 247+ bytes)
            if is_ruuvi_air:
                await self._try_set_ruuvi_air_mtu(client)

            tx_char, rx_char = self._get_history_service_characteristics(client)

            data_queue: asyncio.Queue[bytearray | None] = asyncio.Queue()
            # Buffer for assembling split Ruuvi Air packets
            packet_buffer = bytearray()
            notification_handler = self._create_history_notification_handler(is_ruuvi_air, packet_buffer, data_queue)
            await client.start_notify(tx_char, notification_handler)

            command = self._create_send_history_command(start_time, use_air_format=is_ruuvi_air)

            log.debug("Sending command: %s", command.hex())
            await client.write_gatt_char(rx_char, command)
            log.debug("Sent history command to device")

            async for packet in self._iter_history_queue(data_queue, max_items=max_items):
                yield packet

        except Exception as e:
            log.error("Failed to get history data from device %s: %r", mac, e)
            raise
        finally:
            if client:
                await client.disconnect()
                log.debug("Disconnected from device %s", mac)

    async def _connect_gatt(self, mac: str, max_retries: int = 3) -> BleakClient:
        # Connect to a BLE device using GATT.
        # NOTE: On macOS, the device address is not a MAC address, but a system specific ID
        client = BleakClient(mac)

        for attempt in range(max_retries):
            try:
                await client.connect()
                return client
            except Exception as e:  # noqa: PERF203
                if attempt == max_retries - 1:
                    raise
                log.debug("Connection attempt %s failed: %s - Retrying...", attempt + 1, str(e))
                await asyncio.sleep(1)

        return client  # Satisfy linter - this line will never be reached

    async def _try_set_ruuvi_air_mtu(self, client: BleakClient, mtu: int = 247) -> None:
        """Best-effort MTU negotiation for Ruuvi Air history transfers."""
        try:
            if hasattr(client, "set_mtu"):
                await client.set_mtu(mtu)  # type: ignore[attr-defined]
                log.debug("MTU set to %d", mtu)
        except Exception as e:
            log.debug("Could not set MTU: %s (continuing with default)", e)

    def _create_history_notification_handler(
        self,
        is_ruuvi_air: bool,
        packet_buffer: bytearray,
        queue: asyncio.Queue[bytearray | None],
    ) -> Callable[[BleakGATTCharacteristic, bytearray], None]:
        if is_ruuvi_air:
            return self._create_ruuvi_air_history_notification_handler(packet_buffer, queue)
        return self._create_ruuvitag_history_notification_handler(queue)

    def _create_ruuvi_air_history_notification_handler(
        self, packet_buffer: bytearray, queue: asyncio.Queue[bytearray | None]
    ) -> Callable[[BleakGATTCharacteristic, bytearray], None]:
        def handler(_, data: bytearray) -> None:
            if data and data[0] == 0x05:
                log.debug("Ignoring heartbeat data")
                return

            packet_buffer.extend(data)
            for packet in self._extract_ruuvi_air_history_packets(packet_buffer):
                action, processed_data = self._process_history_notification(packet, True)
                if self._enqueue_history_notification(action, processed_data, queue):
                    packet_buffer.clear()
                    break

        return handler

    def _create_ruuvitag_history_notification_handler(
        self, queue: asyncio.Queue[bytearray | None]
    ) -> Callable[[BleakGATTCharacteristic, bytearray], None]:
        def handler(_, data: bytearray) -> None:
            if data and data[0] == 0x05:
                log.debug("Ignoring heartbeat data")
                return

            action, processed_data = self._process_history_notification(data, False)
            if action == HistoryNotificationAction.IGNORE:
                return

            log.debug("Received data: %s", processed_data)
            self._enqueue_history_notification(action, processed_data, queue)

        return handler

    async def _iter_history_queue(
        self,
        queue: asyncio.Queue[bytearray | None],
        *,
        max_items: int | None,
        timeout_sec: float = 10.0,
    ) -> AsyncGenerator[bytearray, None]:
        items_received = 0
        while True:
            try:
                data = await asyncio.wait_for(queue.get(), timeout=timeout_sec)
            except asyncio.TimeoutError:
                log.error("Timeout waiting for history data")
                return

            if data is None:
                return

            yield data
            items_received += 1
            if max_items and items_received >= max_items:
                return

    @staticmethod
    def _enqueue_history_notification(
        action: HistoryNotificationAction,
        processed_data: bytearray | None,
        queue: asyncio.Queue[bytearray | None],
    ) -> bool:
        """Enqueue processed history data and signal end/error with a None sentinel."""
        if action == HistoryNotificationAction.IGNORE:
            return False
        queue.put_nowait(processed_data)
        if action in (HistoryNotificationAction.END, HistoryNotificationAction.ERROR):
            queue.put_nowait(None)
            return True
        return False

    @staticmethod
    def _extract_ruuvi_air_history_packets(buffer: bytearray) -> list[bytearray]:
        """
        Extract complete Ruuvi Air history packets from a buffer.

        Ruuvi Air multi-record response packet framing (log write, multi-record):
          0..2: 0x3B 0x3B 0x20
          3:    num_records
          4:    record_length (expected 38 bytes)
          5..:  records (num_records * record_length)

        Packets may arrive split across BLE notifications, so we accumulate into a buffer and slice
        out complete frames when enough bytes are present.
        """
        header = b"\x3b\x3b\x20"
        packets: list[bytearray] = []

        while True:
            pos = buffer.find(header)
            if pos == -1:
                # Keep last 2 bytes in case header spans notifications.
                if len(buffer) > 2:
                    del buffer[:-2]
                return packets
            if pos > 0:
                del buffer[:pos]

            if len(buffer) < 5:
                return packets

            num_records = buffer[3]
            record_length = buffer[4]
            if record_length == 0:
                # Invalid; drop one byte and attempt to resync.
                del buffer[0]
                continue

            expected_size = 5 + (num_records * record_length)
            if len(buffer) < expected_size:
                return packets

            packets.append(bytearray(buffer[:expected_size]))
            del buffer[:expected_size]

    def _get_history_service_characteristics(
        self, client: BleakClient
    ) -> tuple[BleakGATTCharacteristic, BleakGATTCharacteristic]:
        # Get the history service
        # https://docs.ruuvi.com/communication/bluetooth-connection/nordic-uart-service-nus
        history_service = next(
            (service for service in client.services if service.uuid.lower() == RUUVI_HISTORY_SERVICE_UUID.lower()),
            None,
        )
        if not history_service:
            raise RuntimeError("History service not found - device may not support history")

        tx_char = history_service.get_characteristic(RUUVI_HISTORY_TX_CHAR_UUID)
        rx_char = history_service.get_characteristic(RUUVI_HISTORY_RX_CHAR_UUID)

        if not tx_char or not rx_char:
            raise RuntimeError("Required characteristics not found")

        return tx_char, rx_char

    def _create_send_history_command(self, start_time: datetime | None = None, use_air_format: bool = False):
        """
        Create the 11-byte log read command.

        Args:
            start_time: Start time for history data. If None, uses 0 (all data).
            try_air_format: If True, use Ruuvi Air format (0x3B 0x3B 0x21),
                           otherwise use RuuviTag format (0x3A 0x3A 0x11).

        Returns:
            bytearray: 11-byte command
        """
        end_time = int(datetime.now(timezone.utc).timestamp())
        start_time_to_use = int(start_time.timestamp()) if start_time else 0

        if use_air_format:
            # Ruuvi Air format: 0x3B 0x3B 0x21 (multi-record read)
            dest = 0x3B
            src = 0x3B
            op = 0x21
        else:
            # RuuviTag format: 0x3A 0x3A 0x11 (single-record read)
            dest = 0x3A
            src = 0x3A
            op = 0x11

        command = bytearray(
            [
                dest,
                src,
                op,
                (end_time >> 24) & 0xFF,  # Current time byte 1 (most significant)
                (end_time >> 16) & 0xFF,  # Current time byte 2
                (end_time >> 8) & 0xFF,  # Current time byte 3
                end_time & 0xFF,  # Current time byte 4
                (start_time_to_use >> 24) & 0xFF,  # Start time byte 1 (most significant)
                (start_time_to_use >> 16) & 0xFF,  # Start time byte 2
                (start_time_to_use >> 8) & 0xFF,  # Start time byte 3
                start_time_to_use & 0xFF,  # Start time byte 4
            ]
        )

        return command

    @staticmethod
    def _process_history_notification(
        data: bytearray, is_ruuvi_air: bool = False
    ) -> tuple[HistoryNotificationAction, bytearray | None]:
        """
        Process history notification data and determine the action to take.

        Args:
            data: Raw notification data from BLE characteristic
            is_ruuvi_air: True if device is Ruuvi Air, False for RuuviTag

        Returns:
            Tuple of (action_type, data):
            - action_type: One of HistoryNotificationAction constants
            - data: The processed data (or None if action is IGNORE)

        Reference:
        - RuuviTag: https://docs.ruuvi.com/communication/bluetooth-connection/nordic-uart-service-nus/log-read
        - Ruuvi Air: https://github.com/ruuvi/docs/blob/8161abb9a08840fceb409aab69d6a7c12d3d5511/communication/bluetooth-connection/nordic-uart-service-nus/read-logged-history-ruuvi-air.md
        """
        action = HistoryNotificationAction.DATA
        processed: bytearray | None = data

        # Ignore heartbeat data that starts with 0x05
        if data and data[0] == 0x05:
            log.debug("Ignoring heartbeat data")
            action = HistoryNotificationAction.IGNORE
            processed = None
        elif is_ruuvi_air:
            # Ruuvi Air protocol
            if len(data) < 5:
                log.debug("Packet too short: %d bytes", len(data))
                action = HistoryNotificationAction.IGNORE
                processed = None
            # Verify packet header (destination=0x3B, source=0x3B, operation=0x20)
            elif data[0] != 0x3B or data[1] != 0x3B or data[2] != 0x20:
                log.debug("Invalid Ruuvi Air packet header: 0x%02X 0x%02X 0x%02X", data[0], data[1], data[2])
                action = HistoryNotificationAction.IGNORE
                processed = None
            # Check for end marker: num_records = 0
            elif data[3] == 0x00:
                log.debug("Received end-of-logs marker (Ruuvi Air)")
                action = HistoryNotificationAction.END
        # RuuviTag protocol
        # Check for error message first (before end marker check)
        # From docs: Error message has header type 0xF0 with payload 0xFF
        # Example: 0x30 30 F0 FF FF FF FF FF FF FF FF
        elif len(data) >= 11 and data[2] == 0xF0:
            log.debug("Device reported error in log reading")
            action = HistoryNotificationAction.ERROR
        # Check for end-of-logs marker (0x3A 0x3A 0x10 0xFF ...)
        # From docs: End marker is 0x3A 3A 10 0xFFFFFFFF FFFFFFFF (all 0xFF)
        # Must check that first 3 bytes match the pattern and rest are 0xFF
        elif (
            len(data) >= 3
            and data[0] == 0x3A
            and data[1] == 0x3A
            and data[2] == 0x10
            and all(b == 0xFF for b in data[3:])
        ):
            log.debug("Received end-of-logs marker (RuuviTag)")
            action = HistoryNotificationAction.END

        return action, processed



from __future__ import annotations

import base64
import logging

from ruuvitag_sensor.ruuvi_types import SensorDataUrl

log = logging.getLogger(__name__)


class UrlDecoder:
    """
    Decodes data from RuuviTag url
    Protocol specification:
    https://github.com/ruuvi/ruuvi-sensor-protocols

    Decoder operations are ported from:
    https://github.com/ruuvi/sensor-protocol-for-eddystone-url/blob/master/index.html

    0:   uint8_t     format;          // (0x02 = realtime sensor readings)
    1:   uint8_t     humidity;        // one lsb is 0.5%
    2-3: uint16_t    temperature;     // Signed 8.8 fixed-point notation.
    4-5: uint16_t    pressure;        // (-50kPa)
    6-7: uint16_t    time;            // seconds (now from reset)

    The bytes for temperature, pressure and time are swapped during the encoding
    """

    def _get_temperature(self, decoded: bytearray) -> float:
        """Return temperature in celsius"""
        temp = (decoded[2] & 127) + decoded[3] / 100
        sign = (decoded[2] >> 7) & 1
        if sign == 0:
            return round(temp, 2)
        return round(-1 * temp, 2)

    def _get_humidity(self, decoded: bytearray) -> float:
        """Return humidity %"""
        return decoded[1] * 0.5

    def _get_pressure(self, decoded: bytearray) -> float:
        """Return air pressure hPa"""
        pressure = ((decoded[4] << 8) + decoded[5]) + 50000
        return pressure / 100

    def decode_data(self, encoded) -> SensorDataUrl | None:
        """
        Decode sensor data.

        Returns:
            dict: Sensor values
        """
        try:
            identifier = None
            data_format = 2
            if len(encoded) > 8:
                data_format = 4
                identifier = encoded[8:]
                encoded = encoded[:8]
            decoded = bytearray(base64.b64decode(encoded, "-_"))  # type: ignore
            return {
                "data_format": data_format,
                "temperature": self._get_temperature(decoded),
                "humidity": self._get_humidity(decoded),
                "pressure": self._get_pressure(decoded),
                "identifier": identifier,
            }
        except Exception:
            log.exception("Encoded value: %s not valid", encoded)
            return None



from typing import Literal, TypedDict

DeviceType = Literal["ruuvitag", "ruuvi_air"]


class SensorDataBase(TypedDict):
    data_format: int | str


class SensorDataUrl(SensorDataBase):
    humidity: float
    temperature: float
    pressure: float
    identifier: str | None


class SensorData3(SensorDataBase):
    humidity: float
    temperature: float
    pressure: float
    acceleration: float
    acceleration_x: int
    acceleration_y: int
    acceleration_z: int
    battery: int


class SensorData5(SensorDataBase):
    humidity: float
    temperature: float
    pressure: float
    acceleration: float
    acceleration_x: int
    acceleration_y: int
    acceleration_z: int
    tx_power: int
    battery: int
    movement_counter: int
    measurement_sequence_number: int
    mac: str
    rssi: int | None


class SensorData6(SensorDataBase):
    humidity: float
    temperature: float
    pressure: float
    pm_2_5: float
    co2: int
    voc: int
    nox: int
    luminosity: float
    measurement_sequence_number: int
    calibration_in_progress: bool
    mac: str


class SensorDataE1(SensorData6):
    pm_1: float
    pm_4: float
    pm_10: float


class SensorHistoryData(TypedDict):
    humidity: float | None
    temperature: float | None
    pressure: float | None
    timestamp: int


class SensorAirHistoryData(TypedDict):
    humidity: float | None
    temperature: float | None
    pressure: float | None
    pm_1: float | None
    pm_2_5: float | None
    pm_4: float | None
    pm_10: float | None
    co2: int | None
    voc: int | None
    nox: int | None
    measurement_sequence_number: int | None
    timestamp: int


SensorData = SensorDataUrl | SensorData3 | SensorData5 | SensorData6 | SensorDataE1

DataFormat = int | str | None
RawSensorData = str | None
DataFormatAndRawSensorData = tuple[DataFormat, RawSensorData]

Mac = str
MacAndSensorData = tuple[Mac, SensorData]

RawData = str
MacAndRawData = tuple[str, str]

ByteData = tuple[int, ...]



from ruuvitag_sensor.data_formats import DataFormats


class TestDataFormats:
    def test_convert_data_valid_data(self):
        test_cases = [
            ("1502010611FF990403651652CAE900080018041C0C8BC6", 3, "03651652CAE900080018041C0C8B"),
            ("1502010611FF990403411540C84AFC72FE2FFFC50B89C6", 3, "03411540C84AFC72FE2FFFC50B89"),
        ]
        for x, data_format, result in test_cases:
            encoded = DataFormats.convert_data(x)
            print(encoded[1])
            assert data_format == encoded[0]
            assert result == encoded[1]

    def test_convert_data_not_valid_binary(self):
        data = b"\x99\x04\x03P\x15]\xceh\xfd\x88\x03\x05\x00\x1b\x0c\x13\x00\x00\x00\x00"
        encoded = DataFormats.convert_data(data)
        assert encoded[0] is None
        assert encoded[1] is None

    def test_convert_data_not_valid(self):
        encoded = DataFormats.convert_data("not_valid")
        assert encoded[0] is None
        assert encoded[1] is None

    def test_get_data_format_3_valid_data(self):
        test_cases = [
            "1502010611FF990403651652CAE900080018041C0C8BC6",
            "1502010611FF990403411540C84AFC72FE2FFFC50B89C6",
        ]
        for x in test_cases:
            encoded = DataFormats._get_data_format_3(x)
            assert encoded is not None

    def test_get_data_format_5_valid_data(self):
        test_cases = [
            "1F0201061BFF990405138A5F92C4F3FFE4FFDC0414C4F6EC29BBE62EB92E73E5BC",
            "1F0201061BFF990405138A5F61C4F0FFE4FFDC0414C5B6EC29B3E62EB92E73E5BC",
        ]
        for x in test_cases:
            encoded = DataFormats._get_data_format_5(x)
            assert encoded is not None

    def test_get_data_format_e1_valid_data(self):
        # Test cases based on Ruuvi Air Data Format E1 specification
        # BLE advertisement structure: [total_len][chunks...][rssi]
        # Format E1 has 40 bytes of payload data
        # Chunk 1: 020106 (BLE flags)
        # Chunk 2: 2B FF9904e1[40 bytes of DFE1 data]
        # Total: 30 (48 bytes) + BC (RSSI)
        test_case = (
            "30020106 2BFF9904e1170c5668c79e0065007004bd11ca00c90a0213e0ac000000decdee110000000000cbb8334c884f BC"
        )
        test_case = test_case.replace(" ", "")  # Remove spaces for clarity
        encoded = DataFormats.convert_data(test_case)
        assert encoded is not None
        assert encoded[0] == "E1"
        assert encoded[1].startswith("e1")

    def test_get_data_format_6_valid_data(self):
        # Test cases based on Ruuvi Air Data Format 6 specification
        # BLE advertisement structure: [total_len][chunks...][rssi]
        # Format 6 has 20 bytes of payload data
        # Chunk 1: 020106 (BLE flags)
        # Chunk 2: 17 FF990406[20 bytes of DF6 data]
        # Total: 1C (28 bytes) + BC (RSSI)
        test_case = "1C020106 17FF990406170C5668C79E007000C90501D9FFCD004C884F BC"
        test_case = test_case.replace(" ", "")  # Remove spaces for clarity
        encoded = DataFormats.convert_data(test_case)
        assert encoded[0] == 6
        assert encoded[1] is not None
        assert encoded[1].startswith("06")

    def test_get_data_format_2and4_valid_data(self):
        test_cases = [
            "1F0201060303AAFE1716AAFE10F6037275752E76692F234248415A414D576F77C9",
            "1F0201060303AAFE1716AAFE10F6037275752E76692F234248415A414D576F77C8",
            "1E0201060303AAFE1616AAFE10EE037275752E76692F23416E4159414D5645CC",
            "1E0201060303AAFE1616AAFE10EE037275752E76692F23416E4159414D5645C9",
        ]
        for x in test_cases:
            encoded = DataFormats._get_data_format_2and4(x)
            assert encoded is not None

    def test_convert_data_too_short_data(self):
        test_case = "1E1107DC00240EE5A9E093F3A3B50100406E0B0952757576692042333634AC04"
        (data_format, data) = DataFormats.convert_data(test_case)
        assert data_format is None
        assert data == ""
